# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module implements stat-archiver's various operations."""

import itertools
import os
import stat
import types
from typing import *

from . import _attrs
from . import _core
from . import _format
from . import _sem
from . import _util

# This is intended to avoid infinite recursion from following symlinks.
# TODO This limit is currently rather pointless as the Python recursion limit is reached first.
_MAX_RECURSION_DEPTH = 2 ** 13

class ArchiveInfo:
    TYPE_UNKNOWN = object()
    TYPE_UNRECOGNIZED = object()
    @property
    def semantics(self) -> _sem.ArchiveSemantics:
        return self.__semantics
    @property
    def n_records(self) -> int:
        return self.__n_records
    @property
    def n_paths(self) -> int:
        return self.__n_paths
    @property
    def n_implicit_dirs(self) -> int:
        return self.__n_implicit_dirs
    @property
    def n_records_by_attr(self) -> Mapping[_attrs.Attribute, int]:
        return self.__n_records_by_attr
    @property
    def n_paths_by_ftype(self) -> Mapping[Any, int]:
        """A mapping from file types to the number of unique paths with that file type.

        The key may also be ``TYPE_UNKNOWN`` for unknown file type, or ``TYPE_UNRECOGNIZED`` for unrecognized file type.
        """
        return self.__n_paths_by_ftype
    @property
    def contains_comments(self) -> bool:
        return self.__contains_comments
    @property
    def normalized(self) -> bool:
        return self.__normalized
    @property
    def n_files(self) -> int:
        return self.__n_paths + self.__n_implicit_dirs
    @property
    def attrs(self) -> Tuple[str, ...]:
        return self.__attrs
    def __init__(self, semantics: _sem.ArchiveSemantics, *args, n_records, n_paths, n_implicit_dirs,
            n_records_by_attr, n_paths_by_ftype, contains_comments, normalized, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(semantics, _sem.ArchiveSemantics):
            raise TypeError()
        self.__semantics = semantics
        self.__n_records = int(n_records)
        self.__n_paths = int(n_paths)
        self.__n_implicit_dirs = int(n_implicit_dirs)
        self.__n_records_by_attr = types.MappingProxyType({
            semantics.attrs[attr]: value for attr, value in dict(n_records_by_attr).items()
        })
        self.__n_paths_by_ftype = types.MappingProxyType(dict(n_paths_by_ftype))
        self.__contains_comments = bool(contains_comments)
        self.__normalized = bool(normalized)
        self.__attrs = tuple(sorted(
            frozenset(attr for attr, value in self.__n_records_by_attr.items() if value != 0),
            key=self.__semantics.attrs.sort_attr_key,
        ))
    def pretty(self) -> str:
        indent = ' ' * 4
        return '\n'.join((
            f'Records: {self.n_records}',
            f'Paths: {self.n_paths}',
            f'Implicit Directories: {self.n_implicit_dirs}',
            f'Total Files: {self.n_files}',
            'Records by Attribute:',
            *(
                f'{indent}{attr.key}: {self.n_records_by_attr[attr]}' for attr in self.attrs
            ),
            'Paths by File Type:',
            *(
                f'{indent}{ftype}: {self.n_paths_by_ftype[ftype]}'
                for ftype in _attrs.ATTR_TYPE.TYPES
                if self.n_paths_by_ftype[ftype] != 0
            ),
            *(
                (f'{indent}unknown: {self.n_paths_by_ftype[__class__.TYPE_UNKNOWN]}',)
                if self.n_paths_by_ftype[__class__.TYPE_UNKNOWN] != 0 else ()
            ),
            *(
                (f'{indent}unrecognized: {self.n_paths_by_ftype[__class__.TYPE_UNRECOGNIZED]}',)
                if self.n_paths_by_ftype[__class__.TYPE_UNRECOGNIZED] != 0 else ()
            ),
            f'Contains Comments: {"true" if self.contains_comments else "false"}',
            f'Normalized: {"true" if self.normalized else "false"}',
            '',
        ))

def getattrs(archive: _sem.ArchiveSink, path: str, attrs: Iterable[Union[str, _attrs.Attribute]], dest_path: bytes, *,
        recursive: bool, missing: str, exclude: Optional[_util.OSPathMask] = None, exclude_topmost: bool = False,
        follow_symlinks: bool = False, max_depth: int = _MAX_RECURSION_DEPTH):
    """Recursively get the attributes ``attrs`` of all files in the directory at path ``path`` and store them in
    ``archive`` under the path ``dest_path``.

    Args:
        archive: The destination archive
        path: The target path on the filesystem
        attrs: The attributes to get, in order
        dest_path: The destination path in this archive under which to store the attribute values
        recursive: Indicates if directories should be walked recursively
        missing: The action to take if a target file does not exist
        exclude: If not ``None`` and a target file's path is in this mask, it is skipped (this applies recursively)
        exclude_topmost: If ``True`` and ``path`` is a directory, its attributes are not retrieved (but it may still
                be traversed recursively)
        follow_symlinks: Indicates if symlinks at or under ``path`` should be followed
        max_depth: The maximum recursion depth (zero to only permit getting the attributes of the file at ``path``)
    Raises:
        _core.ArchivingError
        OSError:
        Exception: If raised by call to ``archive.write_record``
    """
    if not isinstance(archive, _sem.ArchiveSink):
        raise TypeError()
    path = str(path)
    try:
        attrs = [archive.attrs[attr] for attr in attrs]
    except KeyError as e:
        raise ValueError() from e
    archive.attrs.sort_attrs(attrs)
    dest_path = bytes(dest_path)
    assert archive.syntax.validate_path(dest_path)
    recursive = bool(recursive)
    max_depth = int(max_depth)
    if max_depth < 0:
        raise _core.ArchivingError('Maximum recursion depth exceeded' +
                (' (infinite symlink cycle?)' if follow_symlinks else ''))
    if exclude is not None and path in exclude:
        return
    try:
        info = os.lstat(path)
        if follow_symlinks and stat.S_ISLNK(info.st_mode):
            path = str(os.path.realpath(path)) # Note that this may still be a symlink
            if exclude is not None and path in exclude:
                return
            info = os.lstat(path)
    except FileNotFoundError:
        if missing == 'fail':
            raise _core.ArchivingError(f'target file does not exist: {path!r}') from None
        elif missing == 'ignore':
            return
        else:
            raise ValueError()
    is_dir = stat.S_ISDIR(info.st_mode) != 0
    if not is_dir or not exclude_topmost:
        for attr in attrs:
            try:
                value = attr.get_and_serialize(path, (info if attr.may_use_fs else _util.FileInfo(info)))
            except _attrs.AttributeNotApplicableError:
                continue
            archive.write_record(itertools.chain((dest_path, attr), value))
    if is_dir and recursive:
        with os.scandir(path) as dirents:
            dirents = list(dirents)
        dirents.sort(key=lambda dirent: str(dirent.name))
        for dirent in dirents:
            entry_path = archive.syntax.join_paths(dest_path, archive.syntax.serialize_path(dirent.name))
            getattrs(archive, dirent.path, attrs, entry_path, recursive=recursive, missing=missing, exclude=exclude,
                    follow_symlinks=follow_symlinks, max_depth=(max_depth - 1))

def setattrs(archive: _sem.ArchiveSource, path: str, *, missing: str, create_missing_parents: bool,
        exclude: Optional[_util.OSPathMask] = None, follow_symlinks: bool = False):
    """Set all attribute values in ``archive`` to the files at or under ``path`` on the filesystem.

    Args:
        archive: The source archive
        path: The target path on the filesystem
        missing: The action to take if a target file does not exist
        create_missing_parents: Indicates if missing parent directories should be created
        exclude: If not ``None`` and a target file's path is in this mask, it is skipped (this applies recursively)
        follow_symlinks: Indicates if symlinks at or under ``path`` should be followed
    Raises:
        _core.ArchivingError
        _attrs.AttributeFormatError
        OSError
        Exception: If raised by call to ``archive.read_record``
    """
    if not isinstance(archive, _sem.ArchiveSource):
        raise TypeError()
    rootpath = str(path)
    del path
    prev_path = None
    dirty = False
    for record in archive.read_records():
        try:
            source_path = next(record)
            attr = next(record)
        except StopIteration:
            raise ValueError() from None
        attr = archive.attrs[attr]
        path = str(os.path.join(rootpath, archive.syntax.deserialize_path(source_path)))
        if attr.small:
            value = b''.join(record)
            record = iter((value,))
            value = attr.deserialize(value)
        else:
            value = None
        if path != prev_path:
            prev_path = None
            if exclude is not None and path in exclude:
                continue
            try:
                info = os.lstat(path)
                if follow_symlinks and stat.S_ISLNK(info.st_mode):
                    path = str(os.path.realpath(path)) # Note that this may still be a symlink
                    if exclude is not None and path in exclude:
                        continue
                    info = os.lstat(path)
            except FileNotFoundError as e:
                if missing == 'fail':
                    raise _core.ArchivingError(f'target file does not exist: {path!r}') from e
                elif missing == 'ignore':
                    continue
                elif missing == 'create':
                    _create_missing(archive, path, source_path, attr, value, parents=create_missing_parents)
                    info = os.lstat(path)
                else:
                    raise ValueError()
        prev_path = path
        if attr.may_use_fs and (dirty or not isinstance(info, os.stat_result)):
            info = os.lstat(path)
        try:
            new_info, dirty = attr.deserialize_and_set(path, (info if attr.may_use_fs else _util.FileInfo(info)),
                    record)
        except _attrs.AttributeNotApplicableError as e:
            raise ArchivingError(f'attribute {attr.key!r} not applicable to file {path!r}') from e
        if dirty:
            info = os.lstat(path)
        if new_info is not None:
            dirty = _set_common_attrs(path, _util.FileInfo(info), new_info)
            info = new_info

def _create_missing(path: str, source_path: bytes, attr: _attrs.Attribute, value: Any, *, parents: bool):
    if attr == _attrs.ATTR_TYPE:
        assert _attrs.ATTR_TYPE.small
        ftype = value
    elif attr == _attrs.ATTR_MODE:
        assert _attrs.ATTR_MODE.small
        ftype = _attrs.ATTR_TYPE.type_from_mode(value)
    elif attr == _attrs.ATTR_EXECUTABLE:
        ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
    elif attr == _attrs.ATTR_TARGET:
        ftype = _attrs.ATTR_TYPE.TYPE_SYMLINK
    elif attr == _attrs.ATTR_CONTENTS:
        ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
    elif attr in _attrs.STANDARD_HASH_ATTRS:
        ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
    else:
        raise _core.ArchivingError("unable to create missing file: the file's type is not known")
    assert _attrs.ATTR_MODE.small
    mode = value if attr == _attrs.ATTR_MODE else None
    if ftype == _attrs.ATTR_TYPE.TYPE_SYMLINK:
        if attr != _attrs.ATTR_TARGET:
            raise _core.ArchivingError('unable to create missing symlink: target is not known')
        assert _attrs.ATTR_TARGET.small
        os.symlink(value, path)
    elif ftype == _attrs.ATTR_TYPE.TYPE_REGULAR:
        os.mknod(path, mode=((mode if mode is not None else 0o666) | stat.S_IFREG)) # obeys umask
    elif ftype == _attrs.ATTR_TYPE.TYPE_DIRECTORY:
        os.mkdir(path, mode=(mode if mode is not None else 0o777)) # obeys umask
    elif ftype == _attrs.ATTR_TYPE.TYPE_PIPE:
        os.mkfifo(path, mode=(mode if mode is not None else 0o666)) # obeys umask
    else:
        raise _core.ArchivingError(f'unable to create missing file: file type not supported: {ftype!r}')

def _set_common_attrs(path: str, info: _util.FileInfo, new_info: _util.FileInfo) -> bool:
    changed = False
    if new_info.mode != info.mode:
        if new_info.mode & _attrs.ATTR_TYPE.MASK != info.mode & _attrs.ATTR_TYPE.MASK:
            raise _core.ArchivingError('changing file type not permitted')
        os.lchmod(path, new_info.mode)
        changed = True
    if new_info.uid != info.uid or new_info.gid != info.gid:
        os.lchown(path, new_info.uid, new_info.gid)
        changed = True
    if new_info.size != info.size:
        raise _core.ArchivingError('changing file size not permitted')
    if new_info.atime != info.atime or new_info.mtime != info.mtime:
        os.utime(path, ns=(int(new_info.atime * 10 ** 9), int(new_info.mtime * 10 ** 9)), follow_symlinks=False)
        changed = True
    return changed

def process(sources: Iterable[_sem.ArchiveSource], sink: _sem.ArchiveSink, *, sort: bool = False,
        filter_attrs: Optional[Iterable[Union[str, _attrs.Attribute]]] = None,
        exclude: Optional[_format.PathMask] = None):
    """
    Raises:
        _attrs.AttributeFormatError
        Exception: If raised by call to ``read_record`` on any of ``sources`` or call to ``sink.write_record``
    """
    syntax = sink.syntax
    buff = _sem.MemoryArchive(semantics=sink.semantics) if sort else sink
    filter_attrs = set(buff.attrs[attr] for attr in filter_attrs) if filter_attrs is not None else None
    for source in sources:
        if source.syntax != syntax:
            raise ValueError()
        for record in source.read_records():
            try:
                path = next(record)
                attr = next(record)
            except StopIteration:
                raise ValueError()
            if exclude is not None and path in exclude:
                continue
            if filter_attrs is not None and attr in filter_attrs:
                continue
            if attr not in buff.attrs:
                raise ValueError()
            buff.write_record(itertools.chain((path, attr), record))
    if sort:
        for record in buff.read_records():
            sink.write_record(record)

def inspect(parser: _format.RawArchiveParser,
        semantics: _sem.ArchiveSemantics = _sem.STANDARD_SEMANTICS) -> 'ArchiveInfo':
    """
    Raises:
        _attrs.AttributeFormatError
        Exception: If raised by call to ``parser.read_record``
    """
    if not isinstance(parser, _format.RawArchiveParser):
        raise TypeError()
    if not isinstance(semantics, _sem.ArchiveSemantics):
        raise TypeError()
    if parser.syntax != semantics.syntax:
        raise ValueError()
    n_records = 0
    files = _format.PathMap(syntax=semantics.syntax)
    n_implicit_dirs = 0
    n_records_by_attr = {attr: 0 for attr in semantics.attrs}
    ftypes = {ftype: 0 for ftype in _attrs.ATTR_TYPE.TYPES}
    normalized = True
    prev_path_key = None
    prev_attr_key = None
    for record in parser.read_records():
        try:
            path = next(record)
            name = str(next(record))
        except StopIteration:
            raise ValueError()
        try:
            attr = semantics.attrs[name]
        except KeyError:
            raise _sem.ArchiveSemanticsError(f'unrecognized attribute name {name!r}') from None
        n_records += 1
        try:
            ftype = files[path]
        except KeyError:
            ftype = ArchiveInfo.TYPE_UNKNOWN
            n_implicit_dirs += files.set_and_parents(path, ftype, _attrs.ATTR_TYPE.TYPE_DIRECTORY)
        if ftype == ArchiveInfo.TYPE_UNKNOWN:
            if attr == _attrs.ATTR_TYPE:
                ftype = _attrs.ATTR_TYPE.deserialize(b''.join(record))
            elif attr == _attrs.ATTR_MODE:
                try:
                    ftype = _attrs.ATTR_TYPE.type_from_mode(_attrs.ATTR_MODE.deserialize(b''.join(record)))
                except OSError:
                    ftype = ArchiveInfo.TYPE_UNRECOGNIZED
            elif attr == _attrs.ATTR_EXECUTABLE:
                ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
            elif attr == _attrs.ATTR_TARGET:
                ftype = _attrs.ATTR_TYPE.TYPE_SYMLINK
            elif attr == _attrs.ATTR_CONTENTS:
                ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
            elif attr in _attrs.STANDARD_HASH_ATTRS:
                ftype = _attrs.ATTR_TYPE.TYPE_REGULAR
            if ftype != ArchiveInfo.TYPE_UNKNOWN:
                files[path] = ftype
        n_records_by_attr[attr] += 1
        path_key = semantics.syntax.sort_path_key(path)
        attr_key = semantics.attrs.sort_attr_key(name)
        if normalized and (
            attr.key != name or
            parser.seen_comments() or (
                prev_path_key is not None and (
                    prev_path_key > path_key or (
                        prev_path_key == path_key and
                        prev_attr_key >= attr_key
                    )
                )
            )
        ): normalized = False
        prev_path_key = path_key
        prev_attr_key = attr_key
    return ArchiveInfo(
        semantics,
        n_records=n_records,
        n_paths=len(files),
        n_implicit_dirs=n_implicit_dirs,
        n_records_by_attr=n_records_by_attr,
        n_paths_by_ftype={
            ftype: sum(1 for path, ft in files.items() if ft == ftype)
            for ftype in (*_attrs.ATTR_TYPE.TYPES, ArchiveInfo.TYPE_UNKNOWN, ArchiveInfo.TYPE_UNRECOGNIZED)
        },
        contains_comments=parser.seen_comments(),
        normalized=normalized,
    )



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import sys

if not ((3, 6) <= sys.version_info < (4,)):
    raise RuntimeError('Unsupported Python version (Python 3.6+ required)')

from ._attrs import *
from ._core import *
from ._format import *
from ._ops import *
from ._sem import *
from ._util import *
from . import app
from .app import main, run
from . import version



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module implements syntax parsing of the stat-archiver file format."""

from abc import ABCMeta, abstractmethod
import collections
import functools
import os
import re
from typing import *

from . import _core
from . import _util

_OS_PATH_SEP_CHAR = str(os.path.sep)
assert len(_OS_PATH_SEP_CHAR) == 1

class ArchiveFormatError(_core.ArchiveParsingError):
    def __init__(self, *args, line: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__line = int(line) if line is not None else None
    def __str__(self):
        s = super().__str__()
        if self.__line is not None:
            s += f' (line {self.__line})'
        return s

class ArchiveSyntax:
    TEXT_CHARS = bytes(range(0x20, 0x7f))
    @property
    def sep_char(self) -> bytes:
        return self.__sep_char
    @property
    def comment_char(self) -> bytes:
        return self.__comment_char
    @property
    def path_sep_char(self) -> bytes:
        return self.__path_sep_char
    @property
    def escape_char(self) -> bytes:
        return self.__escape_char
    @property
    def special_chars(self) -> bytes:
        return self.__special_chars
    @property
    def escape_sequences(self) -> Mapping[int, bytes]:
        return self.__escape_sequences
    @property
    def escape_sequences_lookup_tree(self) -> Mapping[int, Union[Mapping, int]]:
        return self.__escape_sequences_lookup_tree
    def __init__(self, *args, sep_char: bytes = b':', comment_char: bytes = b'#', path_sep_char: bytes = b'/',
            escape_char: bytes = b'\\', special_chars: bytes = b'', **kwargs):
        super().__init__(*args, **kwargs)
        self.__sep_char = bytes(sep_char)
        if len(self.__sep_char) != 1:
            raise ValueError()
        self.__comment_char = bytes(comment_char)
        if len(self.__comment_char) != 1:
            raise ValueError()
        self.__path_sep_char = bytes(path_sep_char)
        if len(self.__path_sep_char) != 1:
            raise ValueError()
        self.__escape_char = bytes(escape_char)
        if len(self.__escape_char) != 1:
            raise ValueError()
        special_chars = bytes(special_chars)
        self.__special_chars = bytes(sorted(frozenset(
            self.__sep_char + self.__comment_char + self.__path_sep_char + self.__escape_char + special_chars
        )))
        if len(self.__special_chars) != 4 + len(special_chars):
            raise ValueError('duplicate special character')
        for cp in self.__special_chars:
            if cp not in __class__.TEXT_CHARS:
                raise ValueError(f'invalid special character: 0x{hex(cp)[2:].zfill(2)}')
        self.__hash = hash((__class__.__qualname__, self.__sep_char, self.__comment_char, self.__path_sep_char,
                self.__escape_char, self.__special_chars))
        self.__escape_sequences = {
            **{cp: b'x' + hex(cp)[2:].lower().zfill(2).encode() for cp in range(0x00, 0x100)},
            0x00: b'0',
            0x07: b'a',
            0x08: b'b',
            0x09: b't',
            0x0a: b'n',
            0x0b: b'v',
            0x0c: b'f',
            0x0d: b'r',
            ord('\\'): b'\\',
        }
        self.__escape_sequences_lookup_tree = {}
        for cp, seq in self.__escape_sequences.items():
            branch = self.__escape_sequences_lookup_tree
            for seq_cp, next_seq_cp in zip(seq, (*seq[1:], None)):
                if next_seq_cp is None:
                    branch[seq_cp] = cp
                    break
                else:
                    if seq_cp not in branch:
                        branch[seq_cp] = {}
                    branch = branch[seq_cp]
        self.__path_regex = (b'(' + self.__string_regex() + b')(' + re.escape(self.__path_sep_char) + b'(' +
                self.__string_regex() + b'))*')
    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, ArchiveSyntax):
            return False
        if other.__hash != self.__hash:
            return False
        if other.__sep_char != self.__sep_char:
            return False
        if other.__comment_char != self.__comment_char:
            return False
        if other.__path_sep_char != self.__path_sep_char:
            return False
        if other.__escape_char != self.__escape_char:
            return False
        if other.__special_chars != self.__special_chars:
            return False
        return True
    def __hash__(self):
        return self.__hash
    def __string_regex(self, permitted_special_chars: bytes = b'') -> bytes:
        return (b'((' + self.string_char_regex(permitted_special_chars) + b')|(' + re.escape(self.__escape_char) +
                b'([0abtnvfr\\]|x[0-9a-f]{2})))*')
    @functools.lru_cache()
    def string_char_regex(self, permitted_special_chars: bytes = b'') -> bytes:
        return b'[' + re.escape(self.string_chars(permitted_special_chars)) + b']'
    @functools.lru_cache()
    def string_chars_pattern(self, permitted_special_chars: bytes = b'') -> 're.Pattern':
        return re.compile(b'(' + self.string_char_regex(permitted_special_chars) + b')*')
    @functools.lru_cache()
    def is_string_codepoint_lookup_table(self, permitted_special_chars: bytes = b''):
        string_chars = self.string_chars(permitted_special_chars)
        return tuple(cp in string_chars for cp in range(0x00, 0x100))
    @functools.lru_cache()
    def string_chars(self, permitted_special_chars: bytes = b'') -> bytes:
        permitted_special_chars = bytes(permitted_special_chars)
        if self.__escape_char[0] in permitted_special_chars:
            raise ValueError()
        if any(cp not in self.__special_chars for cp in permitted_special_chars):
            raise ValueError()
        return bytes(sorted(
            cp for cp in __class__.TEXT_CHARS
            if cp not in self.__special_chars or cp in permitted_special_chars
        ))
    def escape_string(self, string: bytes, permitted_special_chars: bytes = b'') -> bytes:
        """Eescape characters in ``string`` as necessary."""
        result = bytearray()
        lookup = self.is_string_codepoint_lookup_table(permitted_special_chars)
        for cp in string:
            if lookup[cp]:
                result.append(cp)
            else:
                result.extend(self.escape_char)
                result.extend(self.escape_sequences[cp])
        return bytes(result)
    def unescape_string(self, string: bytes, permitted_special_chars: bytes = b'') -> bytes:
        """Unescape characters in ``string`` as necessary.

        Raises:
            ArchiveFormatError
        """
        string = memoryview(string)
        result = bytearray()
        pos = 0
        while pos < len(string):
            sub = self.string_chars_pattern(permitted_special_chars).match(string, pos).group()
            if len(sub) != 0:
                result.extend(sub)
                pos += len(sub)
                continue
            cp = string[pos]
            if cp == self.escape_char[0]:
                branch = self.escape_sequences_lookup_tree
                for pos in range(pos, len(string)):
                    cp = string[pos]
                    try:
                        branch = branch[cp]
                    except KeyError:
                        raise ArchiveFormatError('invalid escape sequence')
                    if isinstance(branch, int):
                        break
                else:
                    raise ArchiveFormatError('invalid escape sequence')
                result.append(branch)
            else:
                raise ArchiveFormatError(f'invalid character in string: 0x{hex(cp)[2:].zfill(2)}')
            pos += 1
        return bytes(result)
    def validate_path(self, path: bytes) -> bool:
        """Validate a serialized path."""
        path = bytes(path)
        if not re.fullmatch(self.__path_regex, path):
            return False
        if path in (b'', b'.'):
            return True
        for comp in path.split(self.__path_sep_char):
            try:
                comp = self.unescape_string(comp)
            except ArchiveFormatError:
                return False
            try:
                comp = comp.decode('utf-8')
            except UnicodeDecodeError:
                return False
            if comp in (b'', b'.', b'..'):
                return False
        return True
    def serialize_path(self, path: str) -> bytes:
        """Serialize an OS path.

        Raises:
            ValueError: If the path is invalid
        """
        path = str(path)
        if path in ('', '.'):
            return path.encode('utf-8')
        path = self.__path_sep_char.join(
            self.escape_string(comp.encode('utf-8')) for comp in path.split(_OS_PATH_SEP_CHAR)
        )
        if not self.validate_path(path):
            raise ValueError('invalid path')
        return path
    def deserialize_path(self, path: bytes) -> str:
        """Deserialize a path into an OS path.

        Raises:
            ArchiveFormatError
            OSError: If the path was valid but cannot be used on this platform
        """
        path = bytes(path)
        if path in (b'', b'.'):
            return path.decode('utf-8')
        if not self.validate_path(path):
            raise ArchiveFormatError('invalid path')
        try:
            comps = tuple(self.unescape_string(comp).decode('utf-8') for comp in path.split(self.__path_sep_char))
        except UnicodeError:
            raise ArchiveFormatError('invalid path: unicode error') from None
        if any('\x00' in comp or _OS_PATH_SEP_CHAR in comp for comp in comps):
            raise OSError('invalid path for this platform')
        return _OS_PATH_SEP_CHAR.join(comps)
    def join_paths(self, *paths: bytes) -> bytes:
        """Join serialized paths."""
        if all(path == b'.' for path in paths):
            return b'.'
        paths = tuple(bytes(path) for path in paths)
        assert all(self.validate_path(path) for path in paths)
        if len(paths) == 1:
            return paths[0]
        if b'' in paths:
            raise ValueError()
        return self.__path_sep_char.join(path for path in paths if path != b'.')
    def split_path(self, path: bytes) -> List[bytes]:
        if path == b'.':
            return []
        return path.split(self.__path_sep_char)
    def sort_paths(self, paths: List[bytes]):
        """Sort serialized paths in-place."""
        paths.sort(key=self.sort_path_key)
    def sort_path_key(self, path: bytes):
        path = bytes(path)
        return tuple(self.unescape_string(comp).decode('utf-8') for comp in path.split(self.__path_sep_char))
STANDARD_SYNTAX = ArchiveSyntax()

class RawArchive(metaclass=ABCMeta): # abstract
    @property
    def syntax(self) -> 'ArchiveSyntax':
        return self.__syntax
    def __init__(self, *args, syntax: 'ArchiveSyntax' = STANDARD_SYNTAX, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(syntax, ArchiveSyntax):
            raise TypeError()
        self.__syntax = syntax

class RawArchiveSource(RawArchive): # abstract
    @abstractmethod
    def read_record(self) -> Optional[Iterator]:
        """Read a raw record from this source.

        The returned iterator may raise the same exceptions as this method.

        Returns:
            ``None`` if there are no records remaining, or an iterator which yields the record's path, the attribute
            name, then zero or more ``bytes`` objects which, when concatenated together, are the serialized attribute
            value (which may be improperly formatted)
        Raises:
            Exception
        """
        raise NotImplementedError()
    def read_records(self) -> Iterator[Iterator]:
        return iter(self.read_record, None)

class RawArchiveSink(RawArchive): # abstract
    @abstractmethod
    def write_record(self, record: Iterator):
        """Write a raw record to this sink.

        The provided iterator will be completely iterated over before this method returns.

        Args:
            record: An iterator which yields the record's path, then the attribute name, then zero or more ``bytes``
                    objects which, when concatenated together, are the serialized attribute value (which may be
                    improperly formatted)
        Raises:
            Exception
        """
        raise NotImplementedError()

class RawArchiveParser(RawArchiveSource): # concrete
    def __init__(self, source: Iterator[bytes], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__source = source
        self.__buff = bytearray()
        self.__buff_pos = 0
        self.__concurrent_reads = []
        self.__seen_comments = False
        self.__line = 1
    def __enbuffer(self, size: int = 1) -> bool:
        """Read data from the source into the buffer.

        Returns:
            ``True`` if the requested data could be read, ``False`` otherwise
        Raises:
            Exception: If raised by the source
        """
        size = int(size)
        while size > 0:
            for block in self.__source:
                break
            else:
                return False
            if block is None:
                raise ValueError()
            block = bytes(block)
            size -= len(block)
            del self.__buff[:self.__buff_pos]
            self.__buff_pos = 0
            self.__buff.extend(block)
        return True
    def __ensure(self, size: int = 1) -> bool:
        return self.__enbuffer(int(size) - (len(self.__buff) - self.__buff_pos))
    def __finish_concurrent_reads(self):
        for it in self.__concurrent_reads:
            for _ in it:
                pass
        self.__concurrent_reads.clear()
    def __at_end(self) -> bool:
        """
        Returns:
            ``True`` if at the end of the file, ``False`` otherwise
        Raises:
            Exception: If raised by source
        """
        return not self.__ensure()
    def __read_codepoint(self) -> int:
        """
        Raises:
            ArchiveFormatError: If at EOF or raised by source
            Exception: If raised by source
        """
        if not self.__ensure():
            raise ArchiveFormatError('unexpected EOF')
        cp = self.__buff[self.__buff_pos]
        self.__buff_pos += 1
        return cp
    def __peek_codepoint(self) -> int:
        """
        Raises:
            ArchiveFormatError: If at EOF or raised by source
            Exception: If raised by source
        """
        if not self.__ensure():
            raise ArchiveFormatError('unexpected EOF')
        return self.__buff[self.__buff_pos]
    def __read_codepoints(self) -> Iterator[int]:
        while self.__ensure():
            yield self.__read_codepoint()
    def __peek_codepoints(self) -> Iterator[int]:
        while self.__ensure():
            yield self.__peek_codepoint()
    def __read_string(self, permitted_special_chars: bytes = b'') -> bytes:
        return bytes(self.__read_string_codepoints(permitted_special_chars))
    def __read_string_codepoints(self, permitted_special_chars: bytes = b'') -> Iterator[int]:
        lookup = self.syntax.is_string_codepoint_lookup_table(permitted_special_chars)
        while True:
            l = len(self.__buff)
            while self.__buff_pos != l:
                cp = self.__buff[self.__buff_pos]
                if not lookup[cp]:
                    break
                self.__buff_pos += 1
                yield cp
            else:
                if self.__ensure():
                    continue
                else:
                    break
            if cp == self.syntax.escape_char[0]:
                self.__buff_pos += 1
                yield self.__read_escape_sequence()
            else:
                break
    def __read_escape_sequence(self) -> int:
        branch = self.syntax.escape_sequences_lookup_tree
        for cp in self.__read_codepoints():
            try:
                branch = branch[cp]
            except KeyError:
                raise ArchiveFormatError('invalid escape sequence', line=self.__line)
            if isinstance(branch, int):
                return branch
        raise ArchiveFormatError('unexpected EOF')
    def __read_path(self) -> bytes:
        comps = []
        while True:
            comps.append(self.__read_string())
            if comps[-1] in (b'', '.'):
                if len(comps) == 1:
                    return comps[-1]
                raise ArchiveFormatError('invalid path', line=self.__line)
            if comps[-1] == '..':
                raise ArchiveFormatError('invalid path', line=self.__line)
            cp = self.__read_codepoint()
            if cp == self.syntax.sep_char[0]:
                break
            elif cp != self.syntax.path_sep_char[0]:
                raise ArchiveFormatError('invalid path syntax', line=self.__line)
        return self.syntax.path_sep_char.join(comps)
    def __read_name(self) -> str:
        name = self.__read_string(self.syntax.path_sep_char)
        if self.__read_codepoint() != self.syntax.sep_char[0]:
            raise ArchiveFormatError('invalid attribute name syntax', line=self.__line)
        return name.decode('utf-8')
    def __read_comment(self):
        if not self.__at_end() and self.__peek_codepoint() == self.syntax.comment_char[0]:
            self.__seen_comments = True
            while not self.__at_end() and self.__peek_codepoint() != ord('\n'):
                if self.__read_codepoint() not in ArchiveSyntax.TEXT_CHARS:
                    raise ArchiveFormatError('invalid character in comment', line=self.__line)
    def read_record(self) -> Optional[Iterator]:
        """
        Raises:
            ArchiveFormatError: If the format of the archive is invalid or raised by source iterator
            Exception: If raised by source iterator
        """
        self.__finish_concurrent_reads()
        self.__read_record_prefix()
        if self.__at_end():
            return None
        it = self.__read_record_blocks()
        self.__concurrent_reads.append(it)
        return it
    def __read_record_blocks(self) -> Iterator:
        yield self.__read_path()
        yield self.__read_name()
        yield from _util.buffered_iter(self.__read_string_codepoints(self.syntax.sep_char + self.syntax.path_sep_char),
                source_bytes=True)
        self.__read_comment()
        if self.__read_codepoint() != ord('\n'):
            raise ArchiveFormatError('expected LF at end of record', line=self.__line)
        self.__line += 1
    def __read_record_prefix(self):
        while not self.__at_end():
            cp = self.__peek_codepoint()
            if cp == ord('\n'):
                self.__read_codepoint()
                self.__line += 1
            elif cp == self.syntax.comment_char[0]:
                self.__read_comment()
            else:
                break
    def seen_comments(self) -> bool:
        return self.__seen_comments

class RawArchiveComposer(RawArchiveSink): # concrete
    def __init__(self, sink: Callable[[bytes], int], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__sink = sink
        self.__buff = bytearray()
    def __debuffer(self, size: int = _core.IO_BLOCK_SIZE - 1):
        """Flush data to sink until the size of the buffer is no greater than ``size``.

        Raises:
            Exception: If raised by sink
        """
        size = int(size)
        while len(self.__buff) > size:
            del self.__buff[:int(self.__sink(memoryview(self.__buff)))]
    def __write(self, data: bytes):
        self.__buff.extend(data)
        self.__debuffer()
    def write_record(self, record: Iterator, comment: Optional[bytes] = None):
        """
        The write may be buffered.

        Raises:
            Exception: If raised by call to sink
        """
        try:
            path = next(record)
            name = next(record)
        except StopIteration:
            raise ValueError()
        self.__write(path)
        self.__write(self.syntax.sep_char)
        self.__write(self.syntax.escape_string(name.encode('utf-8'), self.syntax.path_sep_char))
        self.__write(self.syntax.sep_char)
        for block in record:
            self.__write(self.syntax.escape_string(block, self.syntax.sep_char + self.syntax.path_sep_char))
        self.write_empty(comment)
    def write_empty(self, comment: Optional[bytes] = None):
        if comment is not None:
            comment = bytes(comment)
            if not all(cp in ArchiveSyntax.TEXT_CHARS for cp in comment):
                raise ValueError()
            self.__write(b'#')
            self.__write(comment)
        self.__write(b'\n')
    def flush(self):
        """Flush all :meth:`written records <write_record>` to the underlying sink.

        Raises:
            Exception: If raised by call to sink
        """
        self.__debuffer(0)

class PathMap(collections.abc.MutableMapping):
    """This class implements a sorted mutable mapping where the keys are serialized paths for a specified syntax."""
    __slots__ = (
        '__syntax',
        '__sort_key',
        '__children',
        '__len',
        '__isset',
        '__value',
    )
    def __init__(self, init=(), *args, syntax: 'ArchiveSyntax' = STANDARD_SYNTAX, _sort_key=None, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(syntax, ArchiveSyntax):
            raise TypeError()
        self.__syntax = syntax
        self.__sort_key = _sort_key if _sort_key is not None else lambda key, value: self.__syntax.sort_path_key(key)
        self.__children = _util.SortableDict(key=self.__sort_key)
        self.__len = 0
        self.__isset = False
        for key, value in collections.OrderedDict(init).items():
            self[key] = value
    def __getitem__(self, path: bytes) -> Any:
        for node in self.__walk(path, allow_create=False):
            pass
        if not node.__isset:
            raise KeyError()
        return node.__value
    def __setitem__(self, path: bytes, value: Any):
        nodes = tuple(self.__walk(path, allow_create=True))
        final = nodes[-1]
        final.__value = value
        if not final.__isset:
            final.__isset = True
            for node in nodes:
                node.__len += 1
    def __delitem__(self, path: bytes):
        nodes = tuple(self.__walk(path, allow_create=False, comps=True))
        final, comp = nodes[-1]
        if not final.__isset:
            raise KeyError()
        del final.__value
        final.__isset = False
        for node, comp in nodes:
            node.__len -= 1
        if len(nodes) > 1 and final.__len == 0:
            node, _ = nodes[-2]
            del node.__children[comp]
    def __contains__(self, path: bytes):
        try:
            self[path]
        except KeyError:
            return False
        else:
            return True
    def __iter__(self) -> Iterator[bytes]:
        for comps in self.__iter_key_comps():
            yield self.__syntax.join_paths(*comps)
    def __len__(self):
        return self.__len
    def set_and_parents(self, path: bytes, value: Any, implicit: Any) -> int:
        count = 0
        prev_nodes = []
        for node in self.__walk(path, allow_create=True):
            prev_nodes.append(node)
            if not node.__isset:
                implied = True
                count += 1
                node.__isset = True
                node.__value = implicit
                for n in prev_nodes:
                    n.__len += 1
            else:
                implied = False
        node.__value = value
        if implied:
            count -= 1
        return count
    def clear(self, path: bytes = b'.'):
        try:
            for node in self.__walk(path, allow_create=False):
                pass
        except KeyError:
            return
        # TODO Memory leak: empty nodes are not pruned
        node.__children.clear()
        node.__len = 0
        if node.__isset:
            node.__isset = False
            del node.__value
    def clear_children(self, path: bytes = b'.'):
        try:
            for node in self.__walk(path, allow_create=False):
                pass
        except KeyError:
            return
        # TODO Memory leak: empty nodes are not pruned
        node.__children.clear()
        node.__len = 1 if node.__isset else 0
    def contains_parent(self, path: bytes) -> bool:
        try:
            for node in self.__walk(path, allow_create=False):
                if node.__isset:
                    return True
        except KeyError:
            pass
        return False
    def contains_child(self, path: bytes) -> bool:
        try:
            for node in self.__walk(path, allow_create=False):
                if len(node) == 0:
                    return False
        except KeyError:
            return False
        return len(node) > 1 or (len(node) == 1 and not node.__isset)
    def firstitem(self) -> Tuple[bytes, Any]:
        item = self.__firstitem()
        if item is None:
            raise KeyError()
        comps, value = item
        return self.__syntax.join_paths(*comps), value
    def __firstitem(self) -> Optional[Tuple[List[bytes], Any]]:
        if self.__isset:
            return [], self.__value
        self.__children.sort()
        for comp, node in self.__children.items():
            item = node.__firstitem()
            if item is None:
                continue
            sub_comps, value = item
            sub_comps.insert(0, comp)
            return sub_comps, value
        return None
    def __iter_key_comps(self) -> Iterator[List[bytes]]:
        if self.__isset:
            yield []
        self.__children.sort()
        for comp, node in self.__children.items():
            for sub_comps in node.__iter_key_comps():
                sub_comps.insert(0, comp)
                yield sub_comps
    def __walk(self, path: bytes, *, allow_create: bool, comps: bool = False):
        node = self
        yield node if not comps else (node, None)
        for comp in self.__syntax.split_path(path):
            child = node.__children.get(comp)
            if child is None:
                if allow_create:
                    child = __class__(syntax=self.__syntax, _sort_key=self.__sort_key)
                    node.__children[comp] = child
                else:
                    raise KeyError()
            node = child
            yield node if not comps else (node, comp)

class PathSet(collections.abc.MutableSet):
    """This class implements a mutable set of serialized paths for a specified syntax."""
    @property
    def proxy(self) -> collections.abc.Set:
        return self.__elems.keys()
    def __init__(self, init=(), *args, syntax: 'ArchiveSyntax' = STANDARD_SYNTAX, **kwargs):
        super().__init__(*args, **kwargs)
        self.__elems = PathMap(syntax=syntax)
        self |= init
    def __contains__(self, path: bytes) -> bool:
        return path in self.__elems
    def __iter__(self) -> Iterator[bytes]:
        return iter(self.__elems)
    def __len__(self) -> int:
        return len(self.__elems)
    def clear(self, *args, **kwargs):
        return self.__elems.clear(*args, **kwargs)
    def clear_children(self, *args, **kwargs):
        return self.__elems.clear_children(*args, **kwargs)
    def contains_parent(self, *args, **kwargs):
        return self.__elems.contains_parent(*args, **kwargs)
    def contains_child(self, *args, **kwargs):
        return self.__elems.contains_child(*args, **kwargs)
    def add(self, path: bytes):
        self.__elems.__setitem__(path, None)
    def discard(self, path: bytes):
        try:
            del self.__elems[path]
        except KeyError:
            pass
    def remove(self, path: bytes):
        del self.__elems[path] # Possible KeyError intentional

class PathMask:
    """This class implements a mask of serialized paths for a specified syntax.

    Paths may be masked recursively and non-recursively.
    """
    @property
    def masked(self) -> collections.abc.Set:
        return self.__masked.proxy
    @property
    def rmasked(self) -> collections.abc.Set:
        return self.__rmasked.proxy
    def __init__(self, masked: Iterable[bytes] = (), rmasked: Iterable[bytes] = (), *args,
            syntax: 'ArchiveSyntax' = STANDARD_SYNTAX, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(syntax, ArchiveSyntax):
            raise TypeError()
        self.__syntax = syntax
        self.__rmasked = PathSet()
        for path in rmasked:
            path = bytes(path)
            assert self.__syntax.validate_path(path)
            if self.__rmasked.contains_parent(path):
                continue
            self.__rmasked.add(path)
        self.__masked = PathSet()
        for path in masked:
            path = bytes(path)
            assert self.__syntax.validate_path(path)
            if self.__rmasked.contains_parent(path):
                continue
            self.__masked.add(path)
    def __contains__(self, path: bytes) -> bool:
        path = bytes(path)
        assert self.__syntax.validate_path(path)
        return path in self.__masked or self.__rmasked.contains_parent(path)



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module contains various utilities used throughout the codebase."""

import collections
import fractions
import os
import threading
from typing import *

from . import _core

class SortableDict(collections.abc.MutableMapping):
    def __init__(self, *args, backing_factory=dict, key=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__backing_factory = backing_factory
        self.__key = key if key is not None else lambda key, value: key
        self.__entries = []
        self.__lookup = self.__backing_factory()
        self.__sorted = True
    def __getitem__(self, key):
        lookup = self.__get_lookup()
        i = lookup[key] # Possible KeyError intentional
        k, v = self.__entries[i]
        return v
    def __setitem__(self, key, value):
        lookup = self.__get_lookup()
        i = lookup.get(key)
        if i is None:
            lookup[key] = len(self.__entries)
            self.__entries.append((key, value))
            self.__sorted = False
        else:
            k, v = self.__entries[i]
            self.__entries[i] = (k, value)
    def __delitem__(self, key):
        lookup = self.__get_lookup()
        i = lookup.pop(key) # Possible KeyError intentional
        k, v = self.__entries.pop(i)
        if i != len(self.__entries):
            self.__lookup = None
        return v
    def __iter__(self):
        for k, v in self.__entries:
            yield k
    def __len__(self):
        return len(self.__entries)
    def getitem_by_index(self, index: int) -> Tuple[Any, Any]:
        return self.__entries[index] # Possible IndexError intentional
    def popitem(self, last: bool = True) -> Tuple[Any, Any]:
        if len(self.__entries) == 0:
            raise KeyError()
        if last:
            k, v = self.__entries.pop(-1)
            if self.__lookup is not None:
                del self.__lookup[k]
        else:
            k, v = self.__entries.pop(0)
            self.__lookup = None
        return k, v
    def clear(self):
        self.__entries.clear()
        self.__lookup = self.__backing_factory()
        self.__sorted = True
    def sort(self, key=None):
        if not self.__sorted:
            if key is None:
                key = self.__key
            self.__entries.sort(key=(lambda entry: key(*entry)))
            self.__lookup = None
            self.__sorted = True
    def __get_lookup(self):
        if self.__lookup is None:
            self.__lookup = self.__backing_factory()
            for i, (key, value) in enumerate(self.__entries):
                self.__lookup[key] = i
        return self.__lookup

class OSPathMask:
    """This class implements a mask of OS paths.

    Paths may be masked recursively and non-recursively.
    """
    @property
    def masked(self) -> frozenset:
        return self.__masked
    @property
    def rmasked(self) -> frozenset:
        return self.__rmasked
    def __init__(self, masked: Iterable[str] = (), rmasked: Iterable[str] = (), *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__rmasked = set()
        for path in rmasked:
            path = os.path.abspath(str(path))
            if any(issubpath(path, p) for p in self.__rmasked):
                continue
            self.__rmasked.add(path)
        self.__rmasked = frozenset(self.__rmasked)
        self.__masked = frozenset(
            os.path.abspath(str(path)) for path in masked
            if not any(issubpath(path, p) for p in self.__rmasked)
        )
    def __contains__(self, path: str) -> bool:
        if not isinstance(path, str):
            raise TypeError()
        path = os.path.abspath(str(path))
        return path in self.__masked or any(issubpath(path, p) for p in self.__rmasked)

def issubpath(sub: str, parent: str, *, strict: bool = False) -> bool:
    sub = os.path.normpath(str(sub))
    parent = os.path.normpath(str(parent))
    return os.path.commonpath((sub, parent)) == parent and (not strict or sub != parent)

class FileInfo:
    @property
    def mode(self) -> int:
        return self.__mode
    @property
    def uid(self) -> int:
        return self.__uid
    @property
    def gid(self) -> int:
        return self.__gid
    @property
    def size(self) -> int:
        return self.__size
    @property
    def atime(self) -> fractions.Fraction:
        return self.__atime
    @property
    def mtime(self) -> fractions.Fraction:
        return self.__mtime
    @property
    def ctime(self) -> fractions.Fraction:
        return self.__ctime
    @property
    def __data(self):
        return (self.__mode, self.__uid, self.__gid, self.__size, self.__atime, self.__mtime, self.__ctime)
    def __init__(self, init=None, *args, mode=None, uid=None, gid=None, size=None, atime=None, mtime=None, ctime=None,
            **kwargs):
        super().__init__(*args, **kwargs)
        if init is not None:
            if isinstance(init, FileInfo):
                mode = init.mode
                uid = init.uid
                gid = init.gid
                size = init.size
                atime = init.atime
                mtime = init.mtime
                ctime = init.ctime
            elif isinstance(init, os.stat_result):
                mode = init.st_mode
                uid = init.st_uid
                gid = init.st_gid
                size = init.st_size
                atime = fractions.Fraction(init.st_atime_ns, 10 ** 9)
                mtime = fractions.Fraction(init.st_mtime_ns, 10 ** 9)
                ctime = fractions.Fraction(init.st_mtime_ns, 10 ** 9)
            else:
                raise ValueError()
        elif None in (mode, uid, gid, size, atime, mtime, ctime):
            raise ValueError()
        self.__mode = int(mode)
        if self.__mode < 0:
            raise ValueError()
        self.__uid = int(uid)
        if self.__uid < 0:
            raise ValueError()
        self.__gid = int(gid)
        if self.__gid < 0:
            raise ValueError()
        self.__size = int(size)
        if self.__size < 0:
            raise ValueError()
        self.__atime = atime if not isinstance(atime, fractions.Fraction) else fractions.Fraction(atime)
        assert fraction_is_decimal(atime)
        self.__mtime = mtime if not isinstance(mtime, fractions.Fraction) else fractions.Fraction(mtime)
        assert fraction_is_decimal(mtime)
        self.__ctime = ctime if not isinstance(ctime, fractions.Fraction) else fractions.Fraction(ctime)
        assert fraction_is_decimal(ctime)
    def __eq__(self, other):
        if not isinstance(other, __class__):
            return NotImplemented
        return self.__data == other.__data
    def __hash__(self):
        return hash((__class__.__qualname__, self.__data))
    def replace(self, mode=None, uid=None, gid=None, size=None, atime=None, mtime=None, ctime=None):
        return __class__(
            mode=(mode if mode is not None else self.__mode),
            uid=(uid if uid is not None else self.__uid),
            gid=(gid if gid is not None else self.__gid),
            size=(size if size is not None else self.__size),
            atime=(atime if atime is not None else self.__atime),
            mtime=(mtime if mtime is not None else self.__mtime),
            ctime=(ctime if ctime is not None else self.__ctime),
        )
    def same_mode(self, *infos: 'FileInfo') -> bool:
        return all(info.__mode == self.__mode for info in infos)
    def same_owners(self, *infos: 'FileInfo') -> bool:
        return all(info.__uid == self.__uid and info.__gid == self.__gid for info in infos)
    def same_size(self, *infos: 'FileInfo') -> bool:
        return all(info.__size == self.__size for info in infos)
    def same_times(self, *infos: 'FileInfo') -> bool:
        return all(
            info.__atime == self.__atime and info.__mtime == self.__mtime and info.__ctime == self.__ctime
            for info in infos
        )

def fraction_is_decimal(value: fractions.Fraction) -> bool:
    if not isinstance(value, fractions.Fraction):
        raise TypeError()
    den = int(value.denominator)
    while den & 1 == 0:
        den >>= 1
    while den % 5 == 0:
        den //= 5
    return den == 1

def fraction_as_decimal(value) -> str:
    if not isinstance(value, fractions.Fraction):
        raise TypeError()
    den = int(value.denominator)
    if den == 1:
        return str(int(value))
    n = 0
    while den & 1 == 0:
        den >>= 1
        n += 1
    m = 0
    while den % 5 == 0:
        den //= 5
        m += 1
    if den != 1:
        raise ValueError()
    del den
    places = max(n, m) # Always > 0
    del n, m
    value *= 2 ** places
    value *= 5 ** places
    value = str(int(value)).zfill(places)
    value = value[:-places] + '.' + value[-places:]
    return value

_UMASK_LOCK = threading.Lock()
_CACHED_UMASK = None
def get_umask(*, clear_cache: bool = False):
    """
    Raises:
        OSError
    """
    # This implementation is thread-safe because variable access is atomic
    if clear_cache:
        clear_umask_cache()
    umask = _CACHED_UMASK
    if umask is None:
        with _UMASK_LOCK:
            umask = os.umask(0o077)
            os.umask(umask)
            _CACHED_UMASK = umask
    return umask
def clear_umask_cache():
    with _UMASK_LOCK:
        _CACHED_UMASK = None

def iter_from_file(file, *, close: bool = False):
    try:
        while True:
            block = file.read(_core.IO_BLOCK_SIZE)
            if block is None:
                raise OSError('read blocking')
            if len(block) == 0:
                break
            yield block
    finally:
        if close:
            file.close()

def buffered_iter(source, *, buff_size: int = _core.IO_BLOCK_SIZE, source_bytes: bool = False):
    buff = bytearray()
    for block in source:
        (buff.append if source_bytes else buff.extend)(block)
        if len(buff) > buff_size:
            yield buff
            buff = bytearray()
    if len(buff) != 0:
        yield buff

def writer_from_file(file):
    def write(block):
        if len(block) == 0:
            return 0
        size = file.write(block)
        if size is None:
            raise OSError('write blocking')
        size = int(size)
        if size == 0:
            raise OSError('end of file while writing')
        return size
    return write



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module implements the functionality of the various attributes supported by stat-archiver."""

from abc import ABCMeta, abstractmethod
import fractions
import hashlib
import os
import re
import stat
import time
import types
from typing import *

from . import _core
from . import _format
from . import _util

class AttributeFormatError(_core.ArchiveParsingError):
    pass

class AttributeNotApplicableError(_core.StatArchiverError):
    pass

class Attribute(metaclass=ABCMeta):
    @property
    def key(self) -> str:
        return self.__key
    @property
    def aliases(self) -> Tuple[str, ...]:
        return self.__aliases
    @property
    def names(self) -> Tuple[str, ...]:
        return self.__names
    @property
    def may_use_fs(self) -> bool:
        return self.__may_use_fs
    @property
    def priority(self) -> int:
        return self.__priority
    @property
    def small(self) -> bool:
        return self.__small
    def __init__(self, key: str, *args, aliases: Iterable[str] = (), may_use_fs: bool, priority: int = 0, small: bool,
            **kwargs):
        super().__init__(*args, **kwargs)
        self.__key = str(key)
        if isinstance(aliases, str):
            raise TypeError()
        self.__aliases = tuple(sorted(str(alias) for alias in aliases))
        self.__names = tuple(sorted((self.__key, *self.__aliases)))
        if '' in self.__names or len(self.__names) != len(frozenset(name.lower() for name in self.__names)):
            raise ValueError()
        self.__may_use_fs = bool(may_use_fs)
        self.__priority = int(priority)
        self.__small = bool(small)
    @abstractmethod
    def serialize(self, value: Any) -> bytes:
        """This method is the inverse of :meth:`deserialize`.

        The value provided must be an acceptable value which would be returned by :meth:`deserialize`.
        """
        raise NotImplementedError()
    @abstractmethod
    def deserialize(self, value: bytes) -> Any:
        """Deserialize a value of this attribute.

        Returns:
            The parsed result of ``value`` as a suitable Python object (an ``int``, ``bytes``, etc.) which is hashable
            and supports equality comparison
        Raises:
            AttributeFormatError
        """
        raise NotImplementedError()
    @abstractmethod
    def get(self, path: str, info: Union[os.stat_result, _util.FileInfo]) -> Any:
        """Get the value of this attribute for the file at the path ``path`` given some information about the file.

        If :attr:`may_use_fs` is true, ``info`` must be a ``os.stat_result`` object; otherwise, it must be a
        :class:`_util.FileInfo` object.

        This method never follows symlinks for the last path component.

        Returns:
            A value suitable for passing to :meth:`serialize`
        Raises:
            AttributeNotApplicableError: If the value of this attribute is not defined for the specified file
            _core.ArchivingError: If there's an error getting the value of this attribute
            OSError: If there's an error interacting with the filesystem
        """
        raise NotImplementedError()
    @abstractmethod
    def set(self, path: str, info: Union[os.stat_result, _util.FileInfo],
            value: Any) -> Tuple[Optional[_util.FileInfo], bool]:
        """Set the value of this attribute for the file at the path ``path`` to the value ``value`` given some
        information about the file.

        A :class:`_util.FileInfo` object may be returned with new values for some file attributes. The file's attributes
        must be updated to these values (excluding ``ctime``) by the caller (unless ``new_info`` is ``None``).

        If :attr:`may_use_fs` is true, ``info`` must be a ``os.stat_result`` object; otherwise, it must be a
        ``FileInfo`` object.

        This method never follows symlinks for the last path component.

        The value provided must be an acceptable value which would be returned by :meth:`deserialize`.

        Returns:
            A tuple of the form ``(new_info, modified)`` where:

            * ``new_info`` is ``None`` or a ``FileInfo`` object (see above).
            * ``modified`` is a boolean which is false if and only if the file on the filesystem was not modified in any
               way; this is always false if :attr:`may_use_fs` is false.
        Raises:
            AttributeNotApplicableError: If this attribute cannot be set for the specified file
            _core.ArchivingError: If there's an error setting the value of this attribute
            OSError: If there's an error interacting with the filesystem
        """
        raise NotImplementedError()
    def get_and_serialize(self, path: str, info: Union[os.stat_result, _util.FileInfo]) -> Iterator[bytes]:
        """
        The returned iterator may raise the same exceptions as this method except for
        :exc:`AttributeNotApplicableError`.

        Raises:
            AttributeNotApplicableError
            _core.ArchivingError
            OSError
        """
        return iter((self.serialize(self.get(path, info)),))
    def deserialize_and_set(self, path: str, info: Union[os.stat_result, _util.FileInfo],
            value: Iterator[bytes]) -> Tuple[Optional[_util.FileInfo], bool]:
        """
        Raises:
            AttributeFormatError
            AttributeNotApplicableError
            _core.ArchivingError
            OSError
        """
        return self.set(path, info, self.deserialize(b''.join(value)))

class PrettyableAttribute(Attribute):
    @abstractmethod
    def pretty(self, value: Any) -> str:
        """Create a "pretty" representation of ``value``.

        The "pretty" representation is lossless, deterministic, and easily human-readable.

        The value provided must be an acceptable value which would be returned by :meth:`Attribute.deserialize`.

        Returns:
            The pretty representation
        """
        raise NotImplementedError()

class AttributeSet:
    @property
    def attrs(self) -> Iterable['Attribute']:
        return self.__attrs
    @property
    def attr_keys(self) -> Iterable[str]:
        return self.__attr_keys
    @property
    def attr_names(self) -> Iterable[str]:
        return self.__attr_names
    @property
    def attrs_by_key(self) -> Mapping[str, 'Attribute']:
        return self.__attrs_by_key
    @property
    def attrs_by_name(self) -> Mapping[str, 'Attribute']:
        return self.__attrs_by_name
    def __init__(self, attrs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        attrs = tuple(attrs)
        if not all(isinstance(attr, Attribute) for attr in attrs):
            raise TypeError()
        self.__attrs = []
        self.__attr_keys = []
        self.__attr_names = []
        self.__attrs_by_key = {}
        self.__attrs_by_name = {}
        for attr in attrs:
            self.__attrs.append(attr)
            self.__attr_keys.append(attr.key)
            self.__attr_names.extend(attr.names)
            self.__attrs_by_key[attr.key] = attr
            for name in attr.names:
                self.__attrs_by_name[name] = attr
        if len(self.__attr_names) != len(frozenset(name.lower() for name in self.__attr_names)):
            raise ValueError('attribute name conflict')
        self.__attrs.sort(key=lambda attr: attr.key)
        self.__attrs = tuple(self.__attrs)
        self.__attr_keys.sort()
        self.__attr_keys = tuple(self.__attr_keys)
        self.__attr_names.sort()
        self.__attr_names = tuple(self.__attr_names)
        self.__attrs_by_key = types.MappingProxyType(self.__attrs_by_key)
        self.__attrs_by_name = types.MappingProxyType(self.__attrs_by_name)
        self.__attrs_by_lower_name = {name.lower(): attr for name, attr in self.__attrs_by_name.items()}
    def __getitem__(self, key: Union[str, 'Attribute']) -> 'Attribute':
        """
        Raises:
            KeyError
        """
        if isinstance(key, Attribute):
            if key not in self:
                raise KeyError()
            return key
        elif isinstance(key, str):
            return self.__attrs_by_lower_name[str(key).lower()] # Possible KeyError intentional
        else:
            raise TypeError()
    def __len__(self) -> int:
        return len(self.__attrs)
    def __iter__(self) -> Iterator['Attribute']:
        return iter(self.__attrs)
    def __contains__(self, item: 'Attribute'):
        return item in self.__attrs
    def sort_attrs(self, attrs: List[Union[str, 'Attribute']]):
        attrs.sort(key=self.sort_attr_key)
    def sort_attr_key(self, attr: Union[str, 'Attribute']) -> Any:
        name = attr if isinstance(attr, Attribute) else str(attr)
        attr = self[name]
        name = attr.key if isinstance(name, Attribute) else name
        return -attr.priority, attr.key, name

STANDARD_ATTRS = []

def _standard_attribute(attr):
    if not isinstance(attr, Attribute):
        attr = attr()
    if not isinstance(attr, Attribute):
        raise TypeError()
    STANDARD_ATTRS.append(attr)
    return attr

class NaturalNumberAttribute(Attribute):
    def serialize(self, value: int) -> bytes:
        value = int(value)
        if value < 0:
            raise ValueError()
        return str(value).encode()
    def deserialize(self, value: bytes) -> int:
        value = bytes(value)
        if not re.fullmatch(br'[0-9]+', value):
            raise AttributeFormatError(f'invalid format for value of attribute {self.name!r}')
        return int(value.decode())

class DecimalAttribute(Attribute):
    def serialize(self, value: fractions.Fraction) -> bytes:
        return _util.fraction_as_decimal(value).encode()
    def deserialize(self, value: bytes) -> fractions.Fraction:
        value = bytes(value)
        if not re.fullmatch(br'-?[0-9]+(\.[0-9]+)?', value):
            raise AttributeFormatError(f'invalid format for value of attribute {self.name!r}')
        return fractions.Fraction(value.decode())

class BytesAttribute(Attribute):
    def __init__(self, *args, size: int, **kwargs):
        self.__size = int(size)
        if self.__size < 0:
            raise ValueError()
        super().__init__(*args, small=(size <= 2 ** 10), **kwargs)
    def serialize(self, value: bytes) -> bytes:
        return bytes(value).hex().lower().encode()
    def deserialize(self, value: bytes) -> bytes:
        value = bytes(value)
        if not re.fullmatch(br'([0-9a-f]{2})*', value):
            raise AttributeFormatError(f'invalid format for value of attribute {self.name!r}')
        return bytes.fromhex(value.decode())

class OctalAttribute(Attribute):
    def __init__(self, *args, mask: int, zfill: int = 0, **kwargs):
        super().__init__(*args, **kwargs)
        self.__mask = int(mask)
        if self.__mask < 0:
            raise ValueError()
        self.__zfill = int(zfill)
        if self.__zfill < 0:
            raise ValueError()
    def serialize(self, value: int) -> bytes:
        value = int(value)
        if value < 0 or value & self.__mask != value:
            raise ValueError()
        return ('0o' + oct(value)[2:].zfill(self.__zfill)).encode()
    def deserialize(self, value: bytes) -> int:
        value = bytes(value)
        if not re.fullmatch(br'0o[0-7]+', value):
            raise AttributeFormatError(f'invalid format for value of attribute {self.name!r}')
        value = int(value[2:].decode(), base=8)
        if value & self.__mask != value:
            raise AttributeFormatError(f'invalid value of attribute {self.name!r}: out of range')
        return value

class BoolAttribute(Attribute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, small=True, **kwargs)
    def serialize(self, value: bool) -> bytes:
        return b'true' if value else b'false'
    def deserialize(self, value: bytes) -> bool:
        if value == b'true':
            return True
        elif value == b'false':
            return False
        else:
            raise AttributeFormatError(f'invalid value of attribute {self.name!r}: expected true or false')

@_standard_attribute
class ATTR_TARGET(Attribute):
    def __init__(self):
        super().__init__('target', may_use_fs=True, priority=100, small=True)
    def serialize(self, value: str) -> bytes:
        return str(value).encode('utf-8')
    def deserialize(self, value: bytes) -> str:
        value = bytes(value)
        try:
            return value.decode('utf-8')
        except ValueError as e:
            raise AttributeFormatError('unicode error') from e
    def get(self, path: str, info: os.stat_result) -> str:
        path = str(path)
        if ATTR_TYPE.type_from_mode(info.st_mode) != ATTR_TYPE.TYPE_SYMLINK:
            raise AttributeNotApplicableError()
        return str(os.readlink(path))
    def set(self, path: str, info: os.stat_result, value: str) -> Tuple[Optional[_util.FileInfo], bool]:
        path = str(path)
        value = str(value)
        if ATTR_TYPE.type_from_mode(info.st_mode) != ATTR_TYPE.TYPE_SYMLINK:
            raise AttributeNotApplicableError()
        target = str(os.readlink(path))
        if target != value:
            try:
                os.remove(path)
            except FileNotFoundError:
                pass
            os.symlink(value, path)
        return None, True

@_standard_attribute
class ATTR_TYPE(Attribute):
    MASK = 0o170000
    TYPES = []
    TYPE_SOCKET = 'socket'
    TYPES.append(TYPE_SOCKET)
    TYPE_SYMLINK = 'symbolic link'
    TYPES.append(TYPE_SYMLINK)
    TYPE_REGULAR = 'regular file'
    TYPES.append(TYPE_REGULAR)
    TYPE_BLOCKDEV = 'block device'
    TYPES.append(TYPE_BLOCKDEV)
    TYPE_DIRECTORY = 'directory'
    TYPES.append(TYPE_DIRECTORY)
    TYPE_CHARDEV = 'character device'
    TYPES.append(TYPE_CHARDEV)
    TYPE_PIPE = 'pipe' # AKA named pipe or FIFO
    TYPES.append(TYPE_PIPE)
    TYPE_DOOR = 'door'
    TYPES.append(TYPE_DOOR)
    TYPE_EVENTPORT = 'event port'
    TYPES.append(TYPE_EVENTPORT)
    TYPE_WHITEOUT = 'whiteout'
    TYPES.append(TYPE_WHITEOUT)
    TYPES = tuple(TYPES)
    TYPES_BY_MODE = {
        0o140000: TYPE_SOCKET,
        0o120000: TYPE_SYMLINK,
        0o100000: TYPE_REGULAR,
        0o060000: TYPE_BLOCKDEV,
        0o040000: TYPE_DIRECTORY,
        0o020000: TYPE_CHARDEV,
        0o010000: TYPE_PIPE,
    }
    assert len(frozenset(TYPES_BY_MODE.values())) == len(TYPES_BY_MODE)
    MODES_BY_TYPE = {typ: mode for mode, typ in TYPES_BY_MODE.items()}
    assert MODES_BY_TYPE[TYPE_SOCKET] ==    stat.S_IFSOCK
    assert MODES_BY_TYPE[TYPE_SYMLINK] ==   stat.S_IFLNK
    assert MODES_BY_TYPE[TYPE_REGULAR] ==   stat.S_IFREG
    assert MODES_BY_TYPE[TYPE_BLOCKDEV] ==  stat.S_IFBLK
    assert MODES_BY_TYPE[TYPE_DIRECTORY] == stat.S_IFDIR
    assert MODES_BY_TYPE[TYPE_CHARDEV] ==   stat.S_IFCHR
    assert MODES_BY_TYPE[TYPE_PIPE] ==      stat.S_IFIFO
    TYPECHARS_BY_TYPE = {
        TYPE_SOCKET:    's',
        TYPE_SYMLINK:   'l',
        TYPE_REGULAR:   '-',
        TYPE_BLOCKDEV:  'b',
        TYPE_DIRECTORY: 'd',
        TYPE_CHARDEV:   'c',
        TYPE_PIPE:      'p',
        TYPE_DOOR:      'D',
        TYPE_EVENTPORT: 'P',
    }
    @staticmethod
    def type_from_mode(mode: int) -> str:
        mode = int(mode)
        if mode < 0:
            raise ValueError()
        ftype = __class__.TYPES_BY_MODE.get(mode & __class__.MASK)
        if ftype is None:
            raise OSError(f'unrecognized file type: 0o{oct(mode)[2:].zfill(6)}')
        return ftype
    @staticmethod
    def mode_from_type(ftype: str) -> int:
        ftype = str(ftype)
        mode = __class__.MODES_BY_TYPE.get(ftype)
        if mode is None:
            if ftype not in __class__.TYPES:
                raise ValueError()
            raise OSError(f'this platform does not support the file type {ftype!r}')
        return mode
    def __init__(self):
        super().__init__('type', may_use_fs=False, priority=95, small=True)
    def serialize(self, value: str) -> bytes:
        value = str(value)
        if value not in __class__.TYPES:
            raise ValueError()
        return value.encode('ascii')
    def deserialize(self, value: bytes) -> str:
        value = bytes(value)
        try:
            value = value.decode('ascii')
        except ValueError:
            raise AttributeFormatError('unrecognized file type') from None
        if value not in __class__.TYPES:
            raise AttributeFormatError(f'unrecognized file type: {value!r}')
        return value
    def get(self, path: str, info: _util.FileInfo) -> str:
        return __class__.type_from_mode(info.mode)
    def set(self, path: str, info: _util.FileInfo, value: str) -> Tuple[Optional[_util.FileInfo], bool]:
        value = __class__.mode_from_type(value)
        return info.replace(mode=(info.mode ^ (info.mode & __class__.MASK) ^ value)), False

@_standard_attribute
class ATTR_MODE(OctalAttribute, PrettyableAttribute):
    MASK = 0o177777
    __RWX_LOOKUP = tuple(
        bytes(cp if (bits & (4 >> i)) != 0 else ord('-') for i, cp in enumerate(b'rwx'))
        for bits in range(0o10)
    )
    __EXTRA_BIT_CHARS = ('SlT', 'sst')
    def __init__(self):
        super().__init__('mode', may_use_fs=False, priority=90, small=True, mask=__class__.MASK, zfill=6)
    def get(self, path: str, info: _util.FileInfo) -> int:
        mode = info.mode
        if mode != mode & __class__.MASK:
            raise OSError(f'unrecognized file mode bits: 0o{oct(mode)[2:].zfill(6)}')
        ftype = mode & ATTR_TYPE.MASK
        if ftype not in ATTR_TYPE.TYPES_BY_MODE:
            raise OSError(f'unrecognized file type: 0o{oct(ftype)[2:].zfill(6)}')
        return mode
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(mode=value), False
    def pretty(self, value: int) -> str:
        value = int(value)
        if value < 0:
            raise ValueError()
        result = ATTR_TYPE.TYPECHARS_BY_TYPE.get(ATTR_TYPE.TYPES_BY_MODE[value & ATTR_TYPE.MASK], '?')
        perms = value & 0o777
        result += ATTR_PERMISSIONS.pretty(perms)
        extra = value & 0o007000
        result = list(result)
        for i in range(3):
            if (extra & 0o004000) != 0:
                result[(i + 1) * 3] = __class__.__EXTRA_BIT_CHARS[(perms & 0o100) >> 6]
            extra <<= 1
            perms <<= 3
        return ''.join(result)

class _PermissionsFlagAttribute(BoolAttribute):
    def __init__(self, *args, mask: int, **kwargs):
        super().__init__(*args, may_use_fs=False, **kwargs)
        self.__mask = int(mask)
    def get(self, path: str, info: _util.FileInfo) -> bool:
        return info.mode & self.__mask != 0
    def set(self, path: str, info: _util.FileInfo, value: bool) -> Tuple[Optional[_util.FileInfo], bool]:
        value = self.__mask ^ (_util.get_umask() & self.__mask) if value else 0
        return info.replace(mode=(info.mode ^ (info.mode & self.__mask) ^ value)), False

@_standard_attribute
class ATTR_SUID(_PermissionsFlagAttribute):
    MASK = 0o004000
    def __init__(self):
        super().__init__('suid', priority=87, mask=__class__.MASK)

@_standard_attribute
class ATTR_SGID(_PermissionsFlagAttribute):
    MASK = 0o002000
    def __init__(self):
        super().__init__('sgid', priority=85, mask=__class__.MASK)

@_standard_attribute
class ATTR_STICKY(_PermissionsFlagAttribute):
    MASK = 0o001000
    def __init__(self):
        super().__init__('sticky', priority=83, mask=__class__.MASK)

@_standard_attribute
class ATTR_PERMISSIONS(OctalAttribute, PrettyableAttribute):
    MASK = 0o000777
    __RWX_LOOKUP = tuple(
        ''.join(ch if (bits & (4 >> i)) != 0 else '-' for i, ch in enumerate('rwx'))
        for bits in range(0o10)
    )
    def __init__(self):
        super().__init__('permissions', aliases=('perms',), may_use_fs=False, priority=80, small=True,
                mask=__class__.MASK, zfill=3)
    def get(self, path: str, info: _util.FileInfo) -> int:
        return info.mode & __class__.MASK
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(mode=(info.mode ^ (info.mode & __class__.MASK) ^ int(value))), False
    def pretty(self, value: int) -> str:
        value = int(value)
        result = []
        for i in range(3):
            result.append(__class__.__RWX_LOOKUP[(value & 0o700) >> 6])
            value <<= 3
        return ''.join(result)
    def unpretty(self, text: str) -> int:
        text = str(text)
        try:
            result = int()
            for i in range(0, 9, 3):
                result <<= 3
                result |= __class__.__RWX_LOOKUP.index(text[i:i + 3])
        except ValueError:
            raise AttributeFormatError(f'invalid format for pretty-format value of attribute {self.name!r}') from None

@_standard_attribute
class ATTR_READABLE(_PermissionsFlagAttribute):
    MASK = 0o000444
    def __init__(self):
        super().__init__('readable', aliases=('read',), priority=77, mask=__class__.MASK)

@_standard_attribute
class ATTR_WRITABLE(_PermissionsFlagAttribute):
    MASK = 0o000222
    def __init__(self):
        super().__init__('writable', aliases=('write',), priority=75, mask=__class__.MASK)

@_standard_attribute
class ATTR_EXECUTABLE(_PermissionsFlagAttribute):
    MASK = 0o000111
    def __init__(self):
        super().__init__('executable', aliases=('exec',), priority=73, mask=__class__.MASK)
    def get(self, path: str, info: _util.FileInfo) -> int:
        if ATTR_TYPE.type_from_mode(info.mode) != ATTR_TYPE.TYPE_REGULAR:
            raise AttributeNotApplicableError()
        return super().get(path, info)
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        if ATTR_TYPE.type_from_mode(info.mode) != ATTR_TYPE.TYPE_REGULAR:
            raise AttributeNotApplicableError()
        return super().set(path, info, value)

@_standard_attribute
class ATTR_UID(NaturalNumberAttribute):
    def __init__(self):
        super().__init__('uid', may_use_fs=False, priority=60, small=True)
    def get(self, path: str, info: _util.FileInfo) -> int:
        return info.uid
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(uid=value), False

@_standard_attribute
class ATTR_GID(NaturalNumberAttribute):
    def __init__(self):
        super().__init__('gid', may_use_fs=False, priority=50, small=True)
    def get(self, path: str, info: _util.FileInfo) -> int:
        return info.gid
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(gid=value), False

class _TimeAttribute(DecimalAttribute, PrettyableAttribute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, small=True, **kwargs)
    def pretty(self, value: fractions.Fraction) -> str:
        value = int(value * 10 ** 9)
        result = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime(value // 10 ** 9))
        sub = 10 ** 9 + (value % 10 ** 9)
        while sub % 10 == 0:
            sub //= 10
        sub = str(sub)[1:]
        if sub != '':
            result += '.' + sub
        result += 'Z'
        return result

STANDARD_TIME_ATTRS = []

@_standard_attribute
class ATTR_ATIME(_TimeAttribute):
    def __init__(self):
        super().__init__('atime', may_use_fs=False, priority=37)
    def get(self, path: str, info: _util.FileInfo) -> fractions.Fraction:
        return info.atime
    def set(self, path: str, info: _util.FileInfo, value: fractions.Fraction,) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(atime=value), False
STANDARD_TIME_ATTRS.append(ATTR_ATIME)

@_standard_attribute
class ATTR_MTIME(_TimeAttribute):
    def __init__(self):
        super().__init__('mtime', may_use_fs=False, priority=35)
    def get(self, path: str, info: _util.FileInfo) -> fractions.Fraction:
        return info.mtime
    def set(self, path: str, info: _util.FileInfo, value: fractions.Fraction) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(mtime=value), False
STANDARD_TIME_ATTRS.append(ATTR_MTIME)

@_standard_attribute
class ATTR_CTIME(_TimeAttribute):
    def __init__(self):
        super().__init__('ctime', may_use_fs=False, priority=33)
    def get(self, path: str, info: _util.FileInfo) -> fractions.Fraction:
        return info.ctime
    def set(self, path: str, info: _util.FileInfo, value: fractions.Fraction) -> Tuple[Optional[_util.FileInfo], bool]:
        return info.replace(ctime=value), False
STANDARD_TIME_ATTRS.append(ATTR_CTIME)

@_standard_attribute
class ATTR_SIZE(NaturalNumberAttribute, PrettyableAttribute):
    __UNITS = ('bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')
    def __init__(self):
        super().__init__('size', may_use_fs=False, priority=10, small=True)
    def get(self, path: str, info: _util.FileInfo) -> int:
        if ATTR_TYPE.type_from_mode(info.mode) not in (ATTR_TYPE.TYPE_SYMLINK, ATTR_TYPE.TYPE_REGULAR):
            raise AttributeNotApplicableError()
        return info.size
    def set(self, path: str, info: _util.FileInfo, value: int) -> Tuple[Optional[_util.FileInfo], bool]:
        if ATTR_TYPE.type_from_mode(info.mode) not in (ATTR_TYPE.TYPE_SYMLINK, ATTR_TYPE.TYPE_REGULAR):
            raise AttributeNotApplicableError()
        return info.replace(size=value), False
    def pretty(self, value: int) -> str:
        value = fractions.Fraction(value)
        steps = 0
        while value / 1024 >= 1 and value.numerator % 2 == 0 and steps + 1 < len(__class__.__UNITS):
            value /= 1024
            steps += 1
        return _util.fraction_as_decimal(value) + ' ' + __class__.__UNITS[steps]

@_standard_attribute
class ATTR_CONTENTS(Attribute):
    def __init__(self):
        super().__init__('contents', may_use_fs=True, priority=-30, small=False)
    def serialize(self, value: bytes) -> bytes:
        return bytes(value)
    def deserialize(self, value: bytes) -> bytes:
        return bytes(value)
    def get(self, path: str, info: os.stat_result) -> bytes:
        return b''.join(self.get_and_serialize(path, info))
    def set(self, path: str, info: os.stat_result, value: bytes) -> Tuple[Optional[_util.FileInfo], bool]:
        return self.deserialize_and_set(path, info, iter((value,)))
    def get_and_serialize(self, path: str, info: os.stat_result) -> Iterator[bytes]:
        if ATTR_TYPE.type_from_mode(info.st_mode) != ATTR_TYPE.TYPE_REGULAR:
            raise AttributeNotApplicableError()
        path = str(path)
        return _util.iter_from_file(open(path, 'rb'), close=True)
    def deserialize_and_set(self, path: str, info: os.stat_result,
            value: Iterator[bytes]) -> Tuple[Optional[_util.FileInfo], bool]:
        if ATTR_TYPE.type_from_mode(info.st_mode) != ATTR_TYPE.TYPE_REGULAR:
            raise AttributeNotApplicableError()
        path = str(path)
        size = 0
        with open(path, 'wb') as f:
            writer = _util.writer_from_file(f)
            for block in value:
                block = memoryview(block)
                while len(block) != 0:
                    s = writer(block)
                    size += s
                    block = block[s:]
        return _util.FileInfo(info).replace(size=size), True

class _HashlibHashAttribute(BytesAttribute):
    def __init__(self, *args, factory, **kwargs):
        self.__factory = factory
        self.__size = int(self.__factory().digest_size)
        super().__init__(*args, may_use_fs=True, size=self.__size, **kwargs)
    def get(self, path: str, info: os.stat_result) -> bytes:
        path = str(path)
        if ATTR_TYPE.type_from_mode(info.st_mode) != ATTR_TYPE.TYPE_REGULAR:
            raise AttributeNotApplicableError()
        h = self.__factory()
        with open(path, 'rb') as f:
            for block in _util.iter_from_file(f):
                h.update(block)
        return h.digest()
    def set(self, path: str, info: os.stat_result, value: bytes) -> Tuple[Optional[_util.FileInfo], bool]:
        value = bytes(value)
        if len(value) != self.__size:
            raise ValueError()
        if self.get(path, info) != value:
            raise _core.ArchivingError('hash does not match')
        return _util.FileInfo(info), True
    def calc_hash(self, data: bytes) -> bytes:
        return self.__factory(data).digest()

STANDARD_HASH_ATTRS = []

@_standard_attribute
class ATTR_SHA2_256(_HashlibHashAttribute):
    def __init__(self):
        super().__init__('SHA2-256', aliases=('SHA-256', 'SHA256'), priority=-60, factory=hashlib.sha256)
STANDARD_HASH_ATTRS.append(ATTR_SHA2_256)

@_standard_attribute
class ATTR_BLAKE2b_512(_HashlibHashAttribute):
    def __init__(self):
        super().__init__('BLAKE2b-512', aliases=('BLAKE2b',), priority=-70, factory=hashlib.blake2b)
STANDARD_HASH_ATTRS.append(ATTR_BLAKE2b_512)

STANDARD_ATTRS = AttributeSet(STANDARD_ATTRS)



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module implements the semantics of the stat-archiver archive format."""

from abc import ABCMeta, abstractmethod
import functools
import itertools
from typing import *

from . import _attrs
from . import _core
from . import _format
from . import _util

class ArchiveSemanticsError(_core.ArchiveParsingError):
    pass

class ArchiveSemantics:
    @property
    def syntax(self) -> _format.ArchiveSyntax:
        return self.__syntax
    @property
    def attrs(self) -> _attrs.AttributeSet:
        return self.__attrs
    def __init__(self, syntax: _format.ArchiveSyntax, attrs: _attrs.AttributeSet, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(syntax, _format.ArchiveSyntax):
            raise TypeError()
        self.__syntax = syntax
        if not isinstance(attrs, _attrs.AttributeSet):
            raise TypeError()
        self.__attrs = attrs
STANDARD_SEMANTICS = ArchiveSemantics(_format.STANDARD_SYNTAX, _attrs.STANDARD_ATTRS)

class Archive(metaclass=ABCMeta): # abstract
    @property
    def semantics(self) -> 'ArchiveSemantics':
        return self.__semantics
    @property
    def syntax(self) -> _format.ArchiveSyntax:
        return self.__semantics.syntax
    @property
    def attrs(self) -> _attrs.AttributeSet:
        return self.__semantics.attrs
    def __init__(self, *args, semantics: 'ArchiveSemantics' = STANDARD_SEMANTICS, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(semantics, ArchiveSemantics):
            raise TypeError()
        self.__semantics = semantics

class ArchiveSource(Archive): # abstract
    @abstractmethod
    def read_record(self) -> Optional[Iterator]:
        """Read a record from this source.

        The returned iterator may raise the same exceptions as this method.

        Returns:
            ``None`` if there are no records remaining, or an iterator which yields the record's path, the attribute,
            then zero or more ``bytes`` objects which, when concatenated together, are the serialized attribute value
            (which may be improperly formatted)
        Raises:
            Exception
        """
        raise NotImplementedError()
    def read_records(self) -> Iterator[Iterator]:
        """Return and iterator which yields the result of :meth:`read_record` until it is ``None``.

        Raises:
            Exception: If raised by the call to ``read_record``
        """
        return iter(self.read_record, None)

class ArchiveSink(Archive): # abstract
    @abstractmethod
    def write_record(self, record: Iterator):
        """Write a record to this sink.

        The provided iterator is completely iterated over before this method returns.

        Args:
            record: An iterable which yields the record's path, then the attribute or attribute name, then zero or more
                    ``bytes`` objects which, when concatenated together, are the serialized attribute value (which MUST
                    be properly formatted)
        Raises:
            Exception
        """
        raise NotImplementedError()

class ArchiveParser(ArchiveSource): # concrete
    def __init__(self, source: Iterator[bytes], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__parser = _format.RawArchiveParser(source, syntax=self.syntax)
    def read_record(self) -> Optional[Iterator]:
        """
        Raises:
            ArchiveFormatError: If the format of the archive is invalid or raised by source iterator
            ArchiveSemanticsError: If the semantics of the archive is invalid or raised by source iterator
            Exception: If raised by source iterator
        """
        record = self.__parser.read_record()
        if record is None:
            return None
        try:
            path = next(record)
            name = str(next(record))
        except StopIteration:
            raise ValueError() from None
        try:
            attr = self.attrs[name]
        except KeyError:
            raise ArchiveSemanticsError(f'unrecognized attribute name {name!r}') from None
        return itertools.chain((path, attr), record)

class ArchiveComposer(ArchiveSink): # concrete
    def __init__(self, sink: Callable[[bytes], int], *args, annotate: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.__annotate = bool(annotate)
        self.__composer = _format.RawArchiveComposer(sink, syntax=self.syntax)
    def write_record(self, record: Iterator):
        """
        The write may be buffered.

        Raises:
            Exception: If raised by call to sink
        """
        try:
            path = next(record)
            attr = next(record)
        except StopIteration:
            raise ValueError()
        attr = self.attrs[attr]
        name = attr.key
        if self.__annotate and isinstance(attr, _attrs.PrettyableAttribute):
            value = b''.join(record)
            record = iter((value,))
            try:
                value = attr.deserialize(value)
            except _attrs.AttributeFormatError as e:
                raise ValueError() from e
            comment = self.syntax.escape_string(attr.pretty(value).encode('utf-8'), self.syntax.sep_char +
                    self.syntax.path_sep_char)
            del value
        else:
            comment = None
        self.__composer.write_record(itertools.chain((path, name), record), comment=comment)
    def flush(self):
        """Flush all :meth:`written records <write_record>` to the underlying sink.

        Raises:
            Exception: If raised by call to sink
        """
        self.__composer.flush()

class MemoryArchive(ArchiveSource, ArchiveSink): # concrete
    """An archive that is buffered entirely in memory.

    Reading from the archive always produces records in the sorted (normalized) order. Writing multiple values for the
    same path and attribute will result in the previous value being overwritten.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__entries = _format.PathMap(syntax=self.syntax)
        self.__entry_sort_key = lambda key, value: self.attrs.sort_attr_key(key)
        self.__entries_iter = None
    def __new_entry(self):
        return _util.SortableDict(key=self.__entry_sort_key)
    def read_record(self) -> Optional[Iterator]:
        """
        Raises:
        """
        if len(self.__entries) == 0:
            return None
        path, entry = self.__entries.firstitem()
        entry.sort()
        attr, record = entry.popitem(last=False)
        if len(entry) == 0:
            del self.__entries[path]
        return iter((path, attr, attr.serialize(record.value)))
    def write_record(self, record: Iterator):
        """
        Raises:
            _attrs.AttributeFormatError: If the attribute value is improperly formatted
        """
        try:
            path = next(record)
            attr = next(record)
        except StopIteration:
            raise ValueError() from None
        path = bytes(path)
        attr = self.attrs[attr]
        value = attr.deserialize(b''.join(record))
        try:
            entry = self.__entries[path]
        except KeyError:
            entry = self.__new_entry()
            self.__entries[path] = entry
        entry[attr] = Record(path, attr, value, semantics=self.semantics)
    def get_record(self, path: bytes, attr: Union[str, _attrs.Attribute]) -> 'Record':
        """
        Raises:
            KeyError
        """
        path = bytes(path)
        attr = self.attrs[attr]
        return self.__records[path][attr] # Possible KeyError x2 intentional
    def set_record(self, path: bytes, attr: Union[str, _attrs.Attribute], value: Any):
        path = bytes(path)
        attr = self.attrs[attr]
        try:
            entry = self.__entries[path]
        except KeyError:
            entry = self.__new_entry()
            self.__entries[path] = entry
        entry[attr] = Record(path, attr, value)
    def iter_records_by_path(self, path: bytes) -> Iterator['Record']:
        try:
            entry = self.__records[path]
        except KeyError:
            return
        yield from entry.values()

@functools.total_ordering
class Record:
    __slots__ = (
        '__semantics',
        '__path',
        '__attr',
        '__value',
        '__hash',
    )
    @property
    def semantics(self) -> 'ArchiveSemantics':
        return self.__semantics
    @property
    def path(self) -> bytes:
        return self.__path
    @property
    def attr(self) -> _attrs.Attribute:
        return self.__attr
    @property
    def value(self) -> Any:
        return self.__value
    def __init__(self, path: bytes, attr: _attrs.Attribute, value: Any, *args,
            semantics: 'ArchiveSemantics' = STANDARD_SEMANTICS, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(semantics, ArchiveSemantics):
            raise TypeError()
        self.__semantics = semantics
        self.__path = bytes(path)
        if not isinstance(attr, _attrs.Attribute):
            raise TypeError()
        if attr not in self.semantics.attrs:
            raise ValueError()
        self.__attr = attr
        self.__value = value
        self.__hash = None
    def __eq__(self, other):
        if not isinstance(other, __class__):
            return False
        if other.__path != self.__path:
            return False
        if other.__attr != self.__attr:
            return False
        if other.__value != self.__value:
            return False
        return True
    def __hash__(self):
        if self.__hash is None:
            self.__hash = hash((__class__.__qualname__, self.__path, self.__attr.key, self.__value))
        return self.__hash
    def __lt__(self, other):
        if not isinstance(other, __class__):
            return NotImplemented
        if self.__path != other.__path:
            self_path = self.__semantics.syntax.sort_path_key(self.__path)
            other_path = self.__semantics.syntax.sort_path_key(other.__path)
            return self_path < other_path
        if self.__attr != other.__attr:
            self_attr = self.__semantics.attrs.sort_attr_key(self.__attr)
            other_attr = self.__semantics.attrs.sort_attr_key(other.__attr)
            return self_attr < other_attr
        if self.__value != other.__value:
            self_value = self.__attr.serialize(self.__value)
            other_value = other.__attr.serialize(other.__value)
            return self_value < other_value
        return False
    def __getitem__(self, key):
        if key == 0:
            return self.__path
        elif key == 1:
            return self.__attr
        elif key == 2:
            return self.__value
        else:
            raise IndexError()
    def __iter__(self):
        yield self.__path
        yield self.__attr
        yield self.__value



#!/usr/bin/env python3

import sys

if not ((3, 6) <= sys.version_info < (4,)):
    raise RuntimeError('Unsupported Python version (Python 3.6+ required)')

import importlib.util

from setuptools import setup

version_spec = importlib.util.spec_from_file_location('version', 'src/statar/version.py')
version = importlib.util.module_from_spec(version_spec)
version_spec.loader.exec_module(version)
del version_spec

setup(
    name='statar',
    version=version.VERSION_STR,
    description='File attribute archiver utility',
    author='Parker Hoyes',
    author_email='contact@parkerhoyes.com',
    url='https://github.com/parkerhoyes/stat-archiver',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: zlib/libpng License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Archiving',
        'Topic :: Utilities',
    ],
    python_requires='>=3.6',
    packages=[
        'statar',
    ],
    package_dir={
        'statar': 'src/statar',
    },
    scripts=[
        'src/scripts/statar',
    ],
    include_package_data=True,
)



# License for stat-archiver, originally found here:
# https://github.com/parkerhoyes/stat-archiver
#
# Copyright (C) 2019 Parker Hoyes <contact@parkerhoyes.com>
#
# This software is provided "as-is", without any express or implied warranty. In
# no event will the authors be held liable for any damages arising from the use of
# this software.
#
# Permission is granted to anyone to use this software for any purpose, including
# commercial applications, and to alter it and redistribute it freely, subject to
# the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not claim
#    that you wrote the original software. If you use this software in a product,
#    an acknowledgment in the product documentation would be appreciated but is
#    not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

"""This module contains the implementation of the front-end command-line utility for stat-archiver."""

import argparse
import logging
import os
import sys

from . import _attrs
from . import _core
from . import _format
from . import _ops
from . import _sem
from . import _util
from . import version

LOGGER = logging.getLogger(__name__)

PROG = 'statar'

def main():
    sys.exit(run(*sys.argv[1:]))

def run(*args):
    return Application().run(args)

class ApplicationExit(Exception):
    def __init__(self, *args, status: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.status = int(status)

class ApplicationSuccess(ApplicationExit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, status=0, **kwargs)

class ApplicationError(ApplicationExit):
    def __init__(self, *args, status: int = 1, **kwargs):
        super().__init__(*args, status=status, **kwargs)

class ApplicationUsageError(ApplicationError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, status=2, **kwargs)

class ApplicationArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ApplicationUsageError(str(message))

class _HelpAction(argparse.Action):
    def __init__(self, option_strings, *, dest, help=None):
        super().__init__(option_strings, dest=dest, nargs=0, help=help)
    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        raise ApplicationSuccess()

class _VersionAction(argparse.Action):
    def __init__(self, option_strings, *, dest, help=None):
        super().__init__(option_strings, dest=dest, nargs=0, help=help)
    def __call__(self, parser, namespace, values, option_string=None):
        print(f'{PROG} {version.VERSION_STR}')
        raise ApplicationSuccess()

ARGPARSER = ApplicationArgumentParser(prog=PROG, allow_abbrev=False, add_help=False, description='''
This utility provides tools for getting, setting, and serializing the attributes of files or of directory trees
recursively.
''')
ARGPARSER.add_argument('-h', '--help', action=_HelpAction, help='Print usage information and exit')
ARGPARSER.add_argument('--version', action=_VersionAction, help='Print version information and exit')
ARGPARSER.add_argument('-v', '--verbosity', choices=('silent', 'critical', 'error', 'warning', 'info', 'debug'),
        default='info', help='Set logging verbosity')
ARGPARSER.add_argument('-C', '--root', default=None, metavar='path',
        help='Change to this directory before performing any operations; this option is NOT order-sensitive, all \
relative paths provided will be relative to this path')
ARGPARSER.add_argument('--profile', default=None, metavar='path',
        help='Profile the tool using cProfile and save the statistics to the specified path (use an empty string for \
stderr); this option is not part of the public API and may be removed or changed at any time')
SUBARGPARSERS = ARGPARSER.add_subparsers(dest='subcommand', title='subcommands')

class _AttrsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not hasattr(namespace, 'attrs') or namespace.attrs is None:
            namespace.attrs = []
        namespace.attrs.append((True, values))

class _NotAttrsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not hasattr(namespace, 'attrs') or namespace.attrs is None:
            namespace.attrs = []
        namespace.attrs.append((False, values))

ARGPARSER_GET = SUBARGPARSERS.add_parser('get', prog=f'{PROG} ... get', allow_abbrev=False, add_help=False,
        help='get attributes', description='''
Get attributes from the specified files and write them to an archive. The resulting archive will be normalized (unless
--annotate is used). If neither --attrs nor --not-attrs is used, the default attributes (mode, uid, gid, mtime) are
assumed.
''')
ARGPARSER_GET.add_argument('-h', '--help', action=_HelpAction, help='Print usage information and exit')
ARGPARSER_GET.add_argument('-o', '--output', default='', metavar='path',
        help='Write archive to this file; use an empty string for stdout (default)')
ARGPARSER_GET.add_argument('-r', '--recursive', action='store_true',
        help='Walk target paths which are directories recursively')
ARGPARSER_GET.add_argument('-a', '--attrs', action=_AttrsAction, metavar='attrs',
        help='A comma-separated list of attributes to get (subsequent usages add attributes, order sensitive with \
--not-attrs)')
ARGPARSER_GET.add_argument('-n', '--not-attrs', action=_NotAttrsAction, metavar='attrs',
        help='A comma-separated list of attributes NOT to get (subsequent usages remove attributes, order sensitive \
with --attrs)')
ARGPARSER_GET.add_argument('-m', '--missing', choices=('fail', 'ignore'), default='fail',
        help='The action to take for target paths which do not exist')
ARGPARSER_GET.add_argument('-x', '--exclude', action='append', metavar='path',
        help="Don't get the attributes of the file at the specified path")
ARGPARSER_GET.add_argument('-X', '--exclude-recursive', action='append', metavar='path',
        help="Don't get the attributes of the children of the directory (or symlink) at the specified path")
ARGPARSER_GET.add_argument('--exclude-topmost', action='store_true',
        help="Don't get the attributes of the topmost directory when recursively descending a directory")
ARGPARSER_GET.add_argument('-L', '--follow-symlinks', action='store_true',
        help='Follow symbolic links (default is to not follow symlinks)')
ARGPARSER_GET.add_argument('-p', '--prefix', metavar='path', default=None,
        help='A path to prefix the results with in the archive')
ARGPARSER_GET.add_argument('-t', '--annotate', action='store_true',
        help='Add comments to some records with more human-readable descriptions of the attribute values')
ARGPARSER_GET.add_argument('targets', nargs='*', metavar='target', help='Target paths')

ARGPARSER_SET = SUBARGPARSERS.add_parser('set', prog=f'{PROG} ... set', allow_abbrev=False, add_help=False,
        help='set attributes', description='''
Read the specified archives, in order, one record at a time, and write each record to the filesystem by setting
attributes (and creating missing files if --missing=create) as necessary.
''')
ARGPARSER_SET.add_argument('-h', '--help', action=_HelpAction, help='Print usage information and exit')
ARGPARSER_SET.add_argument('-m', '--missing', choices=('fail', 'ignore', 'create'), default='fail',
        help='The action to take for target paths which do not exist')
ARGPARSER_SET.add_argument('-p', '--parents', action='store_true',
        help='Create missing parent directories (their permissions will be 0o777 ^ umask); this only has an effect if \
--missing=create')
ARGPARSER_SET.add_argument('-x', '--exclude', action='append', metavar='path',
        help="Don't get the attributes of the file at the specified path")
ARGPARSER_SET.add_argument('-X', '--exclude-recursive', action='append', metavar='path',
        help="Don't get the attributes of the children of the directory (or symlink) at the specified path")
ARGPARSER_SET.add_argument('-L', '--follow-symlinks', action='store_true',
        help='Follow symbolic links (default is to not follow symlinks)')
ARGPARSER_SET.add_argument('inputs', nargs='*', metavar='input',
        help='Read archives from these files; use an empty string for stdin')

ARGPARSER_PROCESS = SUBARGPARSERS.add_parser('process', prog=f'{PROG} ... process', allow_abbrev=False, add_help=False,
        help='process archive(s)', description='''
Read the specified archives, perform the specified transformations on them, then output the resulting archive. Comments
and empty lines are removed. If --sort is specified, the resulting archive is guaranteed to be normalized; otherwise,
the ordering of records in the result is the same as in the input. When using --sort, care should be taken when passing
very large archives as input as they have to be entirely buffered in memory (although, the in-memory representation is
usually much more space efficient than the archive file format). This shouldn't be a problem unless you're dealing with
a very large number of files, or you have low memory available, or the archive contains file contents.
''')
ARGPARSER_PROCESS.add_argument('-h', '--help', action=_HelpAction, help='Print usage information and exit')
ARGPARSER_PROCESS.add_argument('-o', '--output', default='', metavar='path',
        help='Write archive to this file; use an empty string for stdout (default)')
ARGPARSER_PROCESS.add_argument('-s', '--sort', action='store_true',
        help='Sort the records in the archive and remove records for the same path and attribute (the value of the \
last occuring record in the input is used); if specified, the result is guaranteed to be normalized')
ARGPARSER_PROCESS.add_argument('-a', '--attrs', metavar='attrs', default=None,
        help='A comma-separated list of attributes; if specified, only the records which specify these attributes will \
be included in the result')
ARGPARSER_PROCESS.add_argument('-x', '--exclude', action='append', metavar='path',
        help='Filter out records with this path')
ARGPARSER_PROCESS.add_argument('-X', '--exclude-recursive', action='append', metavar='path',
        help='Filter out records whose path is a (non-strict) subpath of this path')
ARGPARSER_PROCESS.add_argument('-t', '--annotate', action='store_true',
        help='Add comments to some records with more human-readable descriptions of the attribute values')
ARGPARSER_PROCESS.add_argument('inputs', nargs='*', metavar='input',
        help='Read archives from these files; use an empty string for stdin')

ARGPARSER_INSPECT = SUBARGPARSERS.add_parser('inspect', prog=f'{PROG} ... inspect', allow_abbrev=False, add_help=False,
        help='inspect archive', description='''
Read the specified archives, then print statistics about them to stdout.
''')
ARGPARSER_INSPECT.add_argument('-h', '--help', action=_HelpAction, help='Print usage information and exit')
ARGPARSER_INSPECT.add_argument('inputs', nargs='*', metavar='input',
        help='Read archives from these files; use an empty string for stdin')

class Application:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    def run(self, args):
        parser = ARGPARSER
        profiler = None
        try:
            logging.captureWarnings(True)
            root_logger = logging.getLogger()
            root_logger.setLevel(logging.INFO)
            logging_stream_handler = logging.StreamHandler()
            logging_stream_handler.setFormatter(ApplicationLoggingFormatter())
            root_logger.addHandler(logging_stream_handler)
            del logging_stream_handler
            args = ARGPARSER.parse_args(tuple(str(arg) for arg in args))
            if args.verbosity == 'silent':
                logging.disable(logging.CRITICAL)
            else:
                root_logger.setLevel({
                    'critical': logging.CRITICAL,
                    'error': logging.ERROR,
                    'warning': logging.WARNING,
                    'info': logging.INFO,
                    'debug': logging.DEBUG,
                }[args.verbosity])
            del root_logger
            if __debug__:
                LOGGER.warning('debug enabled: reduced performance')
            if args.root is not None:
                try:
                    os.chdir(str(args.root))
                except OSError as e:
                    raise ApplicationError(str(e))
            if args.profile is not None:
                try:
                    import cProfile
                except ImportError as e:
                    raise ApplicationError('cannot import cProfile') from e
                if __debug__:
                    LOGGER.warning('profiling with debug enabled')
                pstats_dest = None if args.profile == '' else os.path.abspath(str(args.profile))
                profiler = cProfile.Profile()
                profiler.enable()
            else:
                profiler = None
            if args.subcommand is None:
                raise ApplicationUsageError('no subcommand specified')
            elif args.subcommand == 'get':
                parser = ARGPARSER_GET
                self.__run_get(args)
            elif args.subcommand == 'set':
                parser = ARGPARSER_SET
                self.__run_set(args)
            elif args.subcommand == 'process':
                parser = ARGPARSER_PROCESS
                self.__run_process(args)
            elif args.subcommand == 'inspect':
                parser = ARGPARSER_INSPECT
                self.__run_inspect(args)
            else:
                raise ApplicationUsageError('invalid subcommand')
        except KeyboardInterrupt as e:
            return 3
        except ApplicationError as e:
            LOGGER.critical(str(e))
            LOGGER.debug('fatal error traceback:', exc_info=e)
            status = e.status
        except ApplicationExit as e:
            status = e.status
        else:
            status = 0
        if profiler is not None:
            profiler.disable()
            import io
            import pstats
            s = io.StringIO()
            ps = pstats.Stats(profiler, stream=s)
            del profiler
            ps.strip_dirs()
            ps.sort_stats('tottime')
            ps.print_stats()
            s = s.getvalue()
            del ps
            if pstats_dest is None:
                sys.stderr.write(s)
            else:
                with open(pstats_dest, 'w') as f:
                    f.write(s)
        return status
    def __run_get(self, args):
        if args.output is None:
            raise ValueError()
        output = str(args.output)
        output = None if output == '' else os.path.abspath(output)
        if args.recursive is None:
            raise ValueError()
        recursive = bool(args.recursive)
        arg_attrs = args.attrs if args.attrs is not None else ()
        if len(arg_attrs) == 0:
            arg_attrs = ((True, 'mode,uid,gid,mtime'),)
        attrs = set()
        for add, lst in arg_attrs:
            for attr in lst.split(','):
                attr = attr.strip()
                if attr == '':
                    continue
                try:
                    attr = _attrs.STANDARD_ATTRS[attr]
                except KeyError:
                    raise ApplicationError(f'unrecognized attribute name: {attr!r}')
                if add:
                    attrs.add(attr)
                else:
                    attrs.discard(attr)
        if args.missing is None:
            raise ValueError()
        missing = str(args.missing)
        if missing not in ('fail', 'ignore'):
            raise ValueError()
        exclude = _util.OSPathMask(
            (os.path.abspath(path) for path in args.exclude) if args.exclude is not None else (),
            (os.path.abspath(path) for path in args.exclude_recursive) if args.exclude_recursive is not None else (),
        )
        if args.exclude_topmost is None:
            raise ValueError()
        exclude_topmost = bool(args.exclude_topmost)
        if args.follow_symlinks is None:
            raise ValueError()
        follow_symlinks = bool(args.follow_symlinks)
        prefix = str(args.prefix) if args.prefix is not None else '.'
        try:
            prefix = _format.STANDARD_SYNTAX.serialize_path(prefix)
        except ValueError:
            raise ApplicationError('invalid value for --prefix: invalid path') from None
        if args.annotate is None:
            raise ValueError()
        annotate = bool(args.annotate)
        if args.targets is None:
            raise ValueError()
        if '' in args.targets:
            raise ApplicationError("invalid target path: ''")
        targets = _util.OSPathMask(args.targets) if not recursive else _util.OSPathMask((), args.targets)
        targets = targets.masked if not recursive else targets.rmasked
        if len(targets) == 0:
            LOGGER.warning('no targets')
        target_paths = _format.PathMap()
        try:
            cwd = os.path.abspath(os.getcwd())
        except OSError as e:
            raise ApplicationError(str(e))
        for target in targets:
            dest_path = os.path.relpath(target, cwd)
            if not _util.issubpath(target, cwd):
                raise ApplicationError(f'target outside CWD: {dest_path!r}')
            try:
                dest_path = _format.STANDARD_SYNTAX.serialize_path(dest_path)
            except ValueError:
                raise ApplicationError(f'invalid target path: {dest_path!r}') from None
            dest_path = _format.STANDARD_SYNTAX.join_paths(prefix, dest_path)
            target_paths[dest_path] = target
        try:
            with _get_raw_stdout() if output is None else open(output, 'wb') as output:
                archive = _sem.ArchiveComposer(_util.writer_from_file(output), annotate=annotate)
                for dest_path, target in target_paths.items():
                    _ops.getattrs(archive, target, attrs, dest_path, recursive=recursive, missing=missing,
                            exclude=exclude, exclude_topmost=exclude_topmost, follow_symlinks=follow_symlinks)
                archive.flush()
        except (OSError, _core.StatArchiverError) as e:
            raise ApplicationError(str(e)) from e
    def __run_set(self, args):
        if args.missing is None:
            raise ValueError()
        missing = str(args.missing)
        if missing not in ('fail', 'ignore', 'create'):
            raise ValueError()
        if args.parents is None:
            raise ValueError()
        parents = bool(args.parents)
        exclude = _util.OSPathMask(
            (os.path.abspath(path) for path in args.exclude) if args.exclude is not None else (),
            (os.path.abspath(path) for path in args.exclude_recursive) if args.exclude_recursive is not None else (),
        )
        if args.follow_symlinks is None:
            raise ValueError()
        follow_symlinks = bool(args.follow_symlinks)
        inputs = tuple(os.path.abspath(path) if path != '' else None for path in (str(path) for path in args.inputs))
        if len(inputs) == 0:
            LOGGER.warning('no inputs (use empty string for stdin)')
        if sum(1 for inpt in inputs if inpt is None) > 1:
            raise ApplicationError('cannot use stdin as input multiple times')
        try:
            cwd = os.path.abspath(os.getcwd())
            for inpt in inputs:
                with open(inpt, 'rb') if inpt is not None else _get_raw_stdin() as inpt:
                    archive = _sem.ArchiveParser(_util.iter_from_file(inpt))
                    _ops.setattrs(archive, cwd, missing=missing, create_missing_parents=parents, exclude=exclude,
                            follow_symlinks=follow_symlinks)
        except (OSError, _core.StatArchiverError) as e:
            raise ApplicationError(str(e)) from e
    def __run_process(self, args):
        if args.output is None:
            raise ValueError()
        output = str(args.output)
        output = None if output == '' else os.path.abspath(output)
        if args.sort is None:
            raise ValueError()
        sort = bool(args.sort)
        attrs = args.attrs if args.attrs is not None else ()
        if len(attrs) == 0:
            attrs = None
        else:
            attrs_set = set()
            for add, lst in attrs:
                for attr in lst.split(','):
                    attr = attr.strip()
                    if attr == '':
                        continue
                    try:
                        attr = _attrs.STANDARD_ATTRS[attr]
                    except KeyError:
                        raise ApplicationError(f'unrecognized attribute name: {attr!r}')
                    if add:
                        attrs_set.add(attr)
                    else:
                        attrs_set.discard(attr)
            attrs = attrs_set
            del attrs_set
        exclude = _format.PathSet()
        for path in args.exclude if args.exclude is not None else ():
            try:
                path  = _format.STANDARD_SYNTAX.serialize_path(path)
            except ValueError as e:
                raise ApplicationError(f'invalid path: {path!r}') from e
            exclude.add(path)
        rexclude = _format.PathSet()
        for path in args.exclude_recursive if args.exclude_recursive is not None else ():
            try:
                path  = _format.STANDARD_SYNTAX.serialize_path(path)
            except ValueError as e:
                raise ApplicationError(f'invalid path: {path!r}') from e
            rexclude.add(path)
        exclude = _format.PathMask(exclude, rexclude)
        del rexclude
        if args.annotate is None:
            raise ValueError()
        annotate = bool(args.annotate)
        inputs = tuple(os.path.abspath(path) if path != '' else None for path in (str(path) for path in args.inputs))
        if len(inputs) == 0:
            LOGGER.warning('no inputs (use empty string for stdin)')
        if sum(1 for inpt in inputs if inpt is None) > 1:
            raise ApplicationError('cannot use stdin as input multiple times')
        try:
            cwd = os.path.abspath(os.getcwd())
            with _get_raw_stdout() if output is None else open(output, 'wb') as output:
                output = _sem.ArchiveComposer(_util.writer_from_file(output), annotate=annotate)
                def inpts():
                    for inpt in inputs:
                        with _get_raw_stdin() if inpt is None else open(inpt, 'rb') as inpt:
                            yield _sem.ArchiveParser(_util.iter_from_file(inpt))
                _ops.process(inpts(), output, sort=sort, filter_attrs=attrs, exclude=exclude)
                output.flush()
        except (OSError, _core.StatArchiverError) as e:
            raise ApplicationError(str(e)) from e
    def __run_inspect(self, args):
        inputs = tuple(os.path.abspath(path) if path != '' else None for path in (str(path) for path in args.inputs))
        if len(inputs) == 0:
            LOGGER.warning('no inputs (use empty string for stdin)')
        if sum(1 for inpt in inputs if inpt is None) > 1:
            raise ApplicationError('cannot use stdin as input multiple times')
        try:
            cwd = os.path.abspath(os.getcwd())
            for i, inpt in enumerate(inputs):
                if len(inputs) != 1:
                    sys.stdout.write(('\n' if i != 0 else '') + (inpt if inpt is not None else '/dev/stdin') + '\n')
                with _get_raw_stdin() if inpt is None else open(os.path.abspath(inpt), 'rb') as f:
                    info = _ops.inspect(_format.RawArchiveParser(_util.iter_from_file(f)))
                sys.stdout.write(info.pretty())
                sys.stdout.flush()
        except (OSError, _core.StatArchiverError) as e:
            raise ApplicationError(str(e)) from e

class ApplicationLoggingFormatter(logging.Formatter):
    def format(self, record):
        message = super().format(record)
        level_name = {
            logging.CRITICAL: 'fatal error',
            logging.ERROR: 'error',
            logging.WARNING: 'warning',
            logging.INFO: 'info',
            logging.DEBUG: 'debug',
        }[record.levelno]
        return f'{PROG}: {level_name}: {message}'

def _get_raw_stdin():
    stdin = sys.__stdin__
    if stdin is None:
        raise ApplicationError('no stdin')
    try:
        return stdin.buffer
    except AttributeError:
        raise ApplicationError('no binary stdin') from None

def _get_raw_stdout():
    stdout = sys.__stdout__
    if stdout is None:
        raise ApplicationError('no stdout')
    try:
        return stdout.buffer
    except AttributeError:
        raise ApplicationError('no binary stdout') from None
